import pytest
from pathlib import Path
import os
from .shared import read, write, append
import pickle

try:
    import dppd
    import dppd_plotnine  # noqa: F401

    dp, X = dppd.dppd()
    has_pyggplot = True
except ImportError:
    has_pyggplot = False
    raise ValueError()
    pass


if has_pyggplot:  # noqa C901
    import pandas as pd
    import pypipegraph2 as ppg
    import subprocess

    def magic(filename):
        """See what linux 'file' commando says about that file"""
        if not os.path.exists(filename):
            raise OSError("Does not exists %s" % filename)
        p = subprocess.Popen(["file", filename], stdout=subprocess.PIPE)
        stdout, stderr = p.communicate()
        return stdout

    @pytest.mark.usefixtures("ppg2_per_test")
    class TestPlotJob:
        def test_basic(self):
            def calc():
                return pd.DataFrame(
                    {"X": list(range(0, 100)), "Y": list(range(50, 150))}
                )

            def plot(df):
                return dp(df).p9().add_point("X", "Y")

            def plot2(df):
                p = dp(df).p9().add_point("Y", "X")
                p.width = 5
                p.height = 2
                return p

            of = "out/test.png"
            p, c, t = ppg.PlotJob(of, calc, plot)
            # p.add_fiddle(lambda p: dp(p).scale_x_continuous(trans="log10").pd)
            p.add_another_plot("out/test2.png", plot2)
            ppg.run()
            assert magic(of).find(b"PNG image") != -1
            assert os.path.exists(of + ".tsv")
            assert os.path.exists("cache/out/test.png")
            assert os.path.exists("out/test2.png")
            assert not os.path.exists("cache/out/test2.png")
            assert not os.path.exists("cache/out/test2.png.tsv")

        def test_basic_skip_table(self):
            def calc():
                return pd.DataFrame(
                    {"X": list(range(0, 100)), "Y": list(range(50, 150))}
                )

            def plot(df):
                return dp(df).p9().add_point("X", "Y")

            of = "out/test.png"
            ppg.PlotJob(of, calc, plot, create_table=False)
            ppg.run()
            assert magic(of).find(b"PNG image") != -1
            assert not os.path.exists(of + ".tsv")
            assert os.path.exists("cache/out/test.png")

        def test_basic_return_dict(self):
            def calc():
                return {
                    "A": pd.DataFrame(
                        {"X": list(range(0, 100)), "Y": list(range(50, 150))}
                    )
                }

            def plot(df):
                p = dp(df["A"]).p9().add_point("X", "Y")
                p.width = 5
                p.height = 1
                return p

            of = "out/test.png"
            ppg.PlotJob(of, calc, plot)
            ppg.run()
            assert magic(of).find(b"PNG image") != -1
            assert read(of + ".tsv").find("#A\n") != -1

        def test_basic_return_dict_non_df_raises(self):
            def calc():
                return {
                    "A": pd.DataFrame(
                        {"X": list(range(0, 100)), "Y": list(range(50, 150))}
                    ),
                    "B": "not_a_df",
                }

            def plot(df):
                return dp(df["A"]).p9().add_point("X", "Y")

            of = "out/test.png"
            p, c, t = ppg.PlotJob(of, calc, plot)
            p.height = 1200
            p.width = 800
            with pytest.raises(ppg.JobsFailed):
                ppg.run()
            assert "did not return a DataFrame" in str(
                ppg.global_pipegraph.last_run_result[c[1].job_id].error
            )

        def test_skip_caching(self):
            def calc():
                write("B", "B")
                return pd.DataFrame(
                    {"X": list(range(0, 100)), "Y": list(range(50, 150))}
                )

            def plot(df):
                if not os.path.exists("A"):
                    raise ValueError()
                if not os.path.exists("B"):
                    raise ValueError()

                return dp(df).p9().add_point("X", "Y").pd

            def prep_job(output_filename):
                write("A", "A")

            def plot2(df):
                p = dp(df).p9().add_point("Y", "X")
                p.width = 5
                p.height = 2
                return p

            p = ppg.FileGeneratingJob("A", prep_job)

            # this tests the correct dependency setting on skip_caching
            of = "out/test.png"
            p2, c2, t2 = ppg.PlotJob(
                of, calc, plot, cache_calc=False, render_args={"width": 2, "height": 4}
            )
            p2.depends_on(p)
            t2.depends_on(
                p
            )  # if you don't cache, you have to take care of this yourself
            p2.add_another_plot("out/test2.png", plot2)
            ppg.run()
            assert magic(of).find(b"PNG image") != -1
            assert not os.path.exists("cache/out/test.png")
            assert os.path.exists("out/test.png")
            assert os.path.exists("out/test2.png")

        def xxxtest_redefiniton_and_skip_changes_raises(self):
            def calc():
                return pd.DataFrame(
                    {"X": list(range(0, 100)), "Y": list(range(50, 150))}
                )

            def plot(df):
                return dp(df).p9().add_point("X", "Y")

            of = "out/test.png"
            ppg.PlotJob(of, calc, plot)
            ppg.PlotJob(of, calc, plot)
            with pytest.raises(ppg.JobRedefinitionError):
                ppg.PlotJob(of, calc, plot, render_args={"something": 55})
            # does not remove the jobs though
            ppg.PlotJob(of, calc, plot, cache_calc=False)
            ppg.PlotJob(of, calc, plot, create_table=False)

            ppg.new(run_mode=ppg.RunMode.NOTEBOOK)
            ppg.PlotJob(of, calc, plot)
            ppg.PlotJob(of, calc, plot)
            ppg.PlotJob(of, calc, plot, cache_calc=False)
            ppg.PlotJob(of, calc, plot, create_table=False)
            ppg.PlotJob(of, calc, plot, render_args={"something": 55})

        def xxxtest_pdf(self):
            def calc():
                return pd.DataFrame(
                    {"X": list(range(0, 100)), "Y": list(range(50, 150))}
                )

            def plot(df):
                return dp(df).p9().add_point("X", "Y")

            of = "out/test.pdf"
            ppg.PlotJob(of, calc, plot)
            ppg.run()
            assert magic(of).find(b"PDF document") != -1

        def xxxtest_raises_on_invalid_filename(self):
            def calc():
                return pd.DataFrame(
                    {"X": list(range(0, 100)), "Y": list(range(50, 150))}
                )

            def plot(df):
                return dp(df).p9().add_point("X", "Y")

            of = "out/test.shu"

            def inner():
                ppg.PlotJob(of, calc, plot)

            with pytest.raises(ValueError):
                inner()

        def test_reruns_just_plot_if_plot_changed(self):
            def calc():
                append("out/calc", "A")
                return pd.DataFrame(
                    {"X": list(range(0, 100)), "Y": list(range(50, 150))}
                )

            def plot(df):
                append("out/plot", "B")
                return dp(df).p9().add_point("X", "Y")

            of = "out/test.png"
            ppg.PlotJob(of, calc, plot)
            ppg.run()
            assert magic(of).find(b"PNG image") != -1
            assert read("out/calc") == "A"
            assert read("out/plot") == "B"

            ppg.new()

            def plot2(df):
                append("out/plot", "B")
                return dp(df).p9().add_point("Y", "X")

            ppg.PlotJob(of, calc, plot2)
            ppg.run()
            assert magic(of).find(b"PNG image") != -1
            assert read("out/calc") == "A"
            assert read("out/plot") == "BB"

        def test_no_rerun_if_ignore_code_changes_and_plot_changes(self):
            def calc():
                append("out/calc", "A")
                return pd.DataFrame(
                    {"X": list(range(0, 100)), "Y": list(range(50, 150))}
                )

            def plot(df):
                append("out/plot", "B")
                return dp(df).p9().add_point("X", "Y")

            of = "out/test.png"
            # note that you already need to ignore the function here
            # otherwise, the fact that the function is now *missing*
            # would trigger downstream
            ppg.PlotJob(of, calc, plot, depend_on_function=False)
            ppg.run()
            assert magic(of).find(b"PNG image") != -1
            assert read("out/calc") == "A"
            assert read("out/plot") == "B"

            ppg.new()

            def plot2(df):
                append("out/plot", "B")
                return dp(df).p9().add_point("Y", "X")

            ppg.PlotJob(of, calc, plot2, depend_on_function=False)
            ppg.run()
            assert magic(of).find(b"PNG image") != -1
            assert read("out/calc") == "A"
            assert read("out/plot") == "B"

        def test_reruns_both_if_calc_changed(self):
            def calc():
                append("out/calc", "A")
                return pd.DataFrame(
                    {"X": list(range(0, 100)), "Y": list(range(50, 150))}
                )

            def plot(df):
                append("out/plot", "B")
                return dp(df).p9().add_point("X", "Y")

            of = "out/test.png"
            ppg.PlotJob(of, calc, plot)
            ppg.run()
            assert magic(of).find(b"PNG image") != -1
            assert read("out/calc") == "A"
            assert read("out/plot") == "B"

            ppg.new()

            def calc2():
                append("out/calc", "A")
                x = 5  # noqa: E157,F841
                return pd.DataFrame(
                    {
                        "X": list(range(1, 101)),
                        "Y": list(range(50, 150)),
                    }  # output must really change
                )

            ppg.PlotJob(of, calc2, plot)
            ppg.run()
            assert magic(of).find(b"PNG image") != -1
            assert read("out/calc") == "AA"
            assert read("out/plot") == "BB"

        def test_no_rerun_if_calc_change_but_ignore_codechanges(self):
            def calc():
                append("out/calc", "A")
                return pd.DataFrame(
                    {"X": list(range(0, 100)), "Y": list(range(50, 150))}
                )

            def plot(df):
                append("out/plot", "B")
                return dp(df).p9().add_point("X", "Y")

            of = "out/test.png"
            ppg.PlotJob(of, calc, plot, depend_on_function=False)
            ppg.run()
            assert magic(of).find(b"PNG image") != -1
            assert read("out/calc") == "A"
            assert read("out/plot") == "B"

            ppg.new()

            def calc2():
                append("out/calc", "A")
                x = 5  # noqa: E157,F841
                return pd.DataFrame(
                    {"X": list(range(0, 100)), "Y": list(range(50, 150))}
                )

            ppg.PlotJob(of, calc2, plot, depend_on_function=False)
            ppg.run()
            assert magic(of).find(b"PNG image") != -1
            assert read("out/calc") == "A"
            assert read("out/plot") == "B"

        def test_raises_if_calc_returns_non_df(self):
            def calc():
                return None

            def plot(df):
                append("out/plot", "B")
                return dp(df).p9().add_point("X", "Y")

            of = "out/test.png"
            job, cache_job, table_job = ppg.PlotJob(of, calc, plot)
            with pytest.raises(ppg.JobsFailed):
                ppg.run(print_failures=False)
            print(ppg.global_pipegraph.last_run_result[cache_job[1].job_id].error)
            assert isinstance(
                ppg.global_pipegraph.last_run_result[cache_job[1].job_id].error,
                ppg.JobError,
            )

        def test_raises_if_plot_returns_non_plot(self):
            # import pyggplot
            def calc():
                return pd.DataFrame(
                    {"X": list(range(0, 100)), "Y": list(range(50, 150))}
                )

            def plot(df):
                return None

            of = "out/test.png"
            ppg.PlotJob(of, calc, plot)
            with pytest.raises(ppg.JobsFailed):
                ppg.run()
            print(type(ppg.global_pipegraph.last_run_result[of].error))
            print(repr(ppg.global_pipegraph.last_run_result[of].error))
            assert isinstance(
                ppg.global_pipegraph.last_run_result[of].error, ppg.JobError
            )
            assert "did not return a plot object" in str(
                ppg.global_pipegraph.last_run_result[of].error
            )

        def test_passing_non_function_for_calc(self):
            def inner():
                ppg.PlotJob("out/a", "shu", lambda df: 1)

            with pytest.raises(ValueError):
                inner()

        def test_passing_non_function_for_plot(self):
            def inner():
                ppg.PlotJob("out/a", lambda: 55, "shu")

            with pytest.raises(ValueError):
                inner()

        def test_passing_non_string_as_jobid(self):
            def inner():
                ppg.PlotJob(5, lambda: 1, lambda df: 34)

            with pytest.raises(TypeError):
                inner()

        def test_unpickling_error(self):
            def calc():
                return pd.DataFrame(
                    {"X": list(range(0, 100)), "Y": list(range(50, 150))}
                )

            def plot(df):
                return dp(df).p9().add_point("X", "Y")

            of = "out/test.png"
            p = ppg.PlotJob(of, calc, plot)
            ppg.run()
            ppg.new()
            p = ppg.PlotJob(of, calc, plot)
            with open("cache/out/test.png", "w") as op:
                op.write("no unpickling")
            os.unlink("out/test.png")  # so it reruns
            with pytest.raises(ppg.JobsFailed):
                ppg.run()
            assert not os.path.exists("out/test.png")
            assert isinstance(
                ppg.global_pipegraph.last_run_result[p[1][0].job_id].error.args[0],
                pickle.UnpicklingError,
            )
            assert "Unpickling error in file" in str(
                ppg.global_pipegraph.last_run_result[p[1][0].job_id].error.args[0]
            )

        def test_add_another_not_returning_plot(self):
            def calc():
                return pd.DataFrame(
                    {"X": list(range(0, 100)), "Y": list(range(50, 150))}
                )

            def plot(df):
                return dp(df).p9().add_point("X", "Y")

            def plot2(df):
                return

            of = "out/test.png"
            p, c, t = ppg.PlotJob(of, calc, plot)
            # p.add_fiddle(lambda p: p.scale_x_log10())
            p2 = p.add_another_plot("out/test2.png", plot2)
            with pytest.raises(ppg.JobsFailed):
                ppg.run()
            assert isinstance(
                ppg.global_pipegraph.last_run_result[p2.job_id].error, ppg.JobError
            )

        def test_matplotlib(self):
            import matplotlib.pyplot as plt
            import numpy as np

            def calc():
                return pd.DataFrame({})

            def plot(df):
                # Data for plotting
                t = np.arange(0.0, 2.0, 0.01)
                s = 1 + np.sin(2 * np.pi * t)

                fig, ax = plt.subplots()
                ax.plot(t, s)

                ax.set(
                    xlabel="time (s)",
                    ylabel="voltage (mV)",
                    title="About as simple as it gets, folks",
                )
                ax.grid()
                return fig

            of = "out/test.png"
            p, c, t = ppg.PlotJob(of, calc, plot)
            ppg.run()
            assert Path(of).exists()



# demonstrate runtime logging
import sys
import pypipegraph2 as ppg
import time
from pathlib import Path
import os
import shutil

p = Path("run/slow_running")
if len(sys.argv) <= 1:
    print("Nuking existing", p)
    if p.exists():
        shutil.rmtree(p)
p.mkdir(exist_ok=True, parents=True)
os.chdir(p)


total_runtime = 10

ppg.new(cores=3, run_mode=ppg.RunMode.CONSOLE)


def gen_jobs_stack(name, runtime, total_runtime=total_runtime):
    def inner(of, runtime=runtime):
        for ii in range(runtime):
            time.sleep(1)
        if "_10" in of.name:
            raise ValueError("expected")
        of.write_text(str(of) + "_" + str(runtime))

    out_jobs = []
    for ii in range(0, total_runtime + 1, runtime):
        ppg.FileGeneratingJob(
            f"{name}_{ii}", inner, resources=ppg.Resources.SingleCore
        )
        if out_jobs:
            out_jobs[-1].depends_on(ii)
    return out_jobs


# gen_jobs_stack("1s", 1)
gen_jobs_stack("30s", 20)
# gen_jobs_stack('60s', 60)


def gen_2nd():
    gen_jobs_stack("1sB", 1, 5)


ppg.JobGeneratingJob("genjobs", gen_2nd)

try:
    ppg.run()
except ppg.RunFailed:
    pass



from enum import Enum, auto


class JobOutcome(Enum):
    NotYet = auto()  # default
    Success = auto()
    Skipped = auto()
    Failed = auto()  # todo rename to Failure
    UpstreamFailed = auto()  # rename to UpstreamFailure
    Pruned = auto()


class RunMode(Enum):
    CONSOLE = 1  # certain redefinitions: FatalGraphException, interactive console, ctrl-c does not work
    NOTEBOOK = 2  # certain redefinitions: warning, no interactive console (todo: gui), control-c,/abort works TODO
    NONINTERACTIVE = (
        3  # such as testing, redefinitions like console, but no gui, ctrl-c works TODO
    )

    def is_strict(self):
        return self is RunMode.CONSOLE or self is RunMode.NONINTERACTIVE


class Resources(Enum):
    SingleCore = "SingleCore"
    AllCores = "AllCores"
    # MemoryHog = "MemoryHog"  # todo
    Exclusive = "Exclusive"
    # RateLimited = "RateLimited"  # todo, think web requests
    RunsHere = "RunsHere"  # in this process
    _RaiseInCoreLock = "_RaiseInCoreLock ="  # for testing
    _RaiseInToNumber = "_RaiseInToNumber ="  # for testing

    def is_external(self):
        return self in (
            Resources.SingleCore,
            Resources.AllCores,
            Resources.Exclusive,
        )  # pragma: no cover - used by interactive

    def to_number(self, max_cores):
        if self is Resources.SingleCore:
            return 1
        elif self is Resources.AllCores:
            return max(max_cores - 1, 1)  # never return less than 1 core
        elif self is Resources.Exclusive:
            return max_cores
        elif self is Resources.RunsHere:
            return 1
        elif self is Resources._RaiseInCoreLock:
            return 0  # which the core lock does not like!
        else:
            raise ValueError("Not a Resource with a given number of cores")



from pathlib import Path
import sys
import pytest
import pypipegraph2 as ppg
from .shared import write, read, Dummy, append, counter

global_test = 0


@pytest.mark.usefixtures("ppg2_per_test")
class TestJobs:
    def test_assert_singletonicity_of_jobs(self):
        ppg.new()
        of = "out/a"
        data_to_write = "hello"

        def do_write(of):
            write(of, data_to_write)

        job = ppg.FileGeneratingJob(of, do_write)
        job2 = ppg.FileGeneratingJob(of, do_write)
        assert job is job2  # change from ppg1

    def test_add_job_twice_is_harmless(self):
        job = ppg.FileGeneratingJob("A", lambda of: 5)
        # implicit add
        assert job.job_id in ppg.global_pipegraph.jobs
        assert ppg.global_pipegraph.jobs[job.job_id] is job
        ppg.global_pipegraph.add(job)
        assert job.job_id in ppg.global_pipegraph.jobs
        assert ppg.global_pipegraph.jobs[job.job_id] is job

    def test_redefining_a_jobid_with_different_class_raises(self):
        ppg.new()
        of = "out/a"
        data_to_write = "hello"

        def do_write(of):
            write(of, data_to_write)

        ppg.FileGeneratingJob(of, do_write)

        def load():
            return "shu"

        with pytest.raises(ValueError):
            ppg.DataLoadingJob(of, load)

    def test_raises_on_non_str_job_id(self):
        with pytest.raises(TypeError):
            ppg.FileGeneratingJob(1234, lambda of: None)

    def test_auto_name(self):
        def inner():
            pass

        a = ppg.FunctionInvariant(inner)
        assert a.job_id == "FITestJobs.test_auto_name.<locals>.inner"
        with pytest.raises(TypeError):
            ppg.FunctionInvariant(lambda: 55)
        with pytest.raises(TypeError):
            ppg.FunctionInvariant(None)

    def test_equality_is_identity(self):
        def write_func(of):
            def do_write(of):
                write(of, "do_write done")

            return of, do_write

        ppg.new()
        jobA = ppg.FileGeneratingJob(*write_func("out/a"))
        jobA1 = ppg.FileGeneratingJob(*write_func("out/a"))
        jobB = ppg.FileGeneratingJob(*write_func("out/b"))
        assert jobA is jobA1
        assert jobA == jobA1
        assert not (jobA == jobB)

    def test_has_hash(self):
        ppg.new()
        jobA = ppg.FileGeneratingJob("out/", lambda of: None)
        assert hasattr(jobA, "__hash__")

    def test_repeated_job_definition_and_dependency_callbacks(self, ppg2_per_test):
        def func(of):
            of.write_text("a")

        a = ppg.FileGeneratingJob("a", func)
        a.depends_on(lambda: counter("ac") and None)
        assert len(a.dependency_callbacks) == 1
        assert not Path("ac").exists()
        a = ppg.FileGeneratingJob("a", func)
        a.depends_on(lambda: counter("ac") and counter("bc") and None)  # 2nd dependency
        assert len(a.dependency_callbacks) == 2
        assert not Path("ac").exists()
        ppg.run()
        assert read("ac") == "2"
        assert read("bc") == "1"

    def test_dependency_callback_plus_job(self):
        data = []

        def load_a():
            data.append("a")

        a = ppg.DataLoadingJob("A", load_a)
        b = lambda: ppg.FileGeneratingJob(  # noqa: E731
            "B", lambda of: of.write_text("b")
        )
        c = ppg.FileGeneratingJob(
            "C", lambda of: of.write_text(Path("B").read_text() + data[0])
        )
        c.depends_on(b, a)
        ppg.run()

    def test_data_loading_MultiFile_dowstream(self, job_trace_log):
        # todo :rename this test?
        def tf(ofs):
            counter("A")
            ofs[0].write_text("a1")
            ofs[1].write_text("a1")

        a = ppg.MultiTempFileGeneratingJob(["a1", "a2"], tf)

        def write(ofs):
            counter("B")
            ofs[0].write_text("b")
            ofs[1].write_text("c" + read("a1"))

        bc = ppg.MultiFileGeneratingJob(["b", "c"], write)
        assert bc[0] == Path("b")
        assert bc[1] == Path("c")
        bc.depends_on(a)
        bc()
        assert read("c") == "ca1"
        assert read("b") == "b"
        assert read("b") == "b"
        assert read("A") == "1"
        assert read("B") == "1"
        ppg.run()
        assert read("c") == "ca1"
        assert read("b") == "b"
        assert read("B") == "1"
        assert read("A") == "1"

    def test_returning_none(self):
        ppg.new(run_mode=ppg.RunMode.NOTEBOOK)
        a = ppg.FileGeneratingJob("out/a", lambda of: counter("a") and write(of, "A"))
        b = ppg.FileGeneratingJob("out/b", lambda of: counter("b") and write(of, "B"))
        c = ppg.DataLoadingJob("o", lambda: counter("c") and None)
        b.depends_on(a)
        a.depends_on_params("x")
        a.depends_on(c)
        b.depends_on(c)
        ppg.run()
        assert read("a") == "1"
        assert read("b") == "1"
        assert read("c") == "1"
        a.depends_on_params("y")
        ppg.run()
        assert read("a") == "2"
        assert read("c") == "2"
        assert read("b") == "1"  # ppg2_rust change


@pytest.mark.usefixtures("ppg2_per_test")
class TestJobs2:
    def test_str(self):
        a = ppg.FileGeneratingJob("out/A", lambda of: write("out/A", "hello"))
        assert isinstance(str(a), str)

        a = ppg.ParameterInvariant("out/A", "hello")
        assert isinstance(str(a), str)

        a = ppg.JobGeneratingJob("out/Ax", lambda: "hello")
        assert isinstance(str(a), str)


@pytest.mark.usefixtures("create_out_dir")
@pytest.mark.usefixtures("ppg2_per_test")
class TestFileGeneratingJob:
    def test_basic(self):
        of = "out/a"
        data_to_write = "hello"

        def do_write(ofof):
            print("do_write was called")
            write(of, data_to_write)

        job = ppg.FileGeneratingJob(of, do_write, depend_on_function=False)
        res = ppg.run()
        assert res[job.job_id].outcome == ppg.enums.JobOutcome.Success
        assert Path(of).exists()
        assert read(of) == data_to_write

    def test_cores_needed(self):
        ppg.new(cores=5)
        of = "out/a"
        data_to_write = "hello"

        def do_write(ofof):
            print("do_write was called")
            write(of, data_to_write)

        job = ppg.FileGeneratingJob(of, do_write, depend_on_function=False)
        with pytest.raises(TypeError):
            job.use_resources(0)
        job.cores_needed = ppg.Resources.SingleCore
        job.cores_needed = ppg.Resources.AllCores
        job.cores_needed = ppg.Resources.Exclusive

        for i in range(10):
            j = ppg.FileGeneratingJob("out/%i" % i, lambda of, i=i: write(of, f"b{i}"))
            if i % 2 == 0:
                j.cores_needed = ppg.Resources.Exclusive

        ppg.run()
        assert read(of) == data_to_write
        for i in range(10):
            assert read(f"out/{i}") == f"b{i}"

    def test_cores_needed2(self):
        ppg.new(cores=3)
        # this is meant to trigger the
        # "this job needed to much resources, or was not runnable"
        # case of graph.start_jobs

        for i in range(20):
            j = ppg.FileGeneratingJob("out/%i" % i, lambda of, i=i: write(of, f"b{i}"))
            if i % 2 == 0:
                j.use_resources(ppg.Resources.AllCores)
            else:
                j.use_resources(ppg.Resources.Exclusive)

        ppg.run()
        for i in range(20):
            assert read(f"out/{i}") == f"b{i}"

    def test_misspecified_job_does_not_hang_graph(self):
        # this occurred during development
        ppg.new(cores=3)

        for i in range(1):
            j = ppg.FileGeneratingJob(
                "out/%i" % i,
                lambda i=i: write("out/%i" % i, "b"),
                depend_on_function=False,
            )

        with pytest.raises(ppg.JobsFailed):
            ppg.run()
        print(
            "this is it'", str(ppg.global_pipegraph.last_run_result["out/0"].error), "'"
        )

        assert "number is required, not PosixPath" in str(
            ppg.global_pipegraph.last_run_result["out/0"].error
        )

    def test_job_not_creating_its_file(self):
        # this occurred during development
        ppg.new(cores=3)

        for i in range(1):
            j = ppg.FileGeneratingJob(
                "out/%i" % i,
                lambda of, i=i: write("shu", "b"),
                depend_on_function=False,
            )

        with pytest.raises(ppg.JobsFailed):
            ppg.run()
        assert "did not create the following files: ['out/0']" in str(
            ppg.global_pipegraph.last_run_result["out/0"].error
        )

    def test_basic_with_parameter(self):
        data_to_write = "hello"

        def do_write(filename):
            print("do_write was called")
            write(filename, data_to_write)

        job = ppg.FileGeneratingJob("out/a", do_write, depend_on_function=False)
        ppg.run()
        assert Path("out/a").exists()
        op = open("out/a", "r")
        data = op.read()
        op.close()
        assert data == data_to_write

    def test_simple_filegeneration_with_function_dependency(self):
        of = "out/a"
        data_to_write = "hello"

        def do_write(ofof):
            print("do_write was called")
            write(of, data_to_write)

        job = ppg.FileGeneratingJob(of, do_write)
        ppg.run()
        assert Path(of).exists()
        op = open(of, "r")
        data = op.read()
        op.close()
        assert data == data_to_write

    def test_filejob_raises_if_no_data_is_written(self):
        of = "out/a"

        def do_write(of):
            write("out/A", "")

        job = ppg.FileGeneratingJob(of, do_write)

        with pytest.raises(ppg.JobsFailed):
            ppg.run()

        assert isinstance(
            ppg.global_pipegraph.last_run_result[job.job_id].error.args[0],
            ppg.JobContractError,
        )
        assert not (Path(of).exists())

    def test_filejob_empty_allowed(self):
        ppg.FileGeneratingJob("out/A", lambda of: write("out/A", ""), empty_ok=True)
        ppg.run()
        assert read("out/A") == ""

    def test_filejob_raising_exception_bubbles(self):
        def do(of):
            raise ValueError("Hello Exception")

        job = ppg.FileGeneratingJob("out/A", do)
        with pytest.raises(ppg.JobsFailed):
            ppg.run()
        assert "Hello Exception" in str(
            ppg.global_pipegraph.last_run_result[job.job_id].error
        )

    def test_simple_filegeneration_keeps_file_on_exception(self):
        of = "out/a"
        data_to_write = "hello"

        def do_write(of):
            write(of, data_to_write)
            raise ValueError("shu")

        ppg.FileGeneratingJob(of, do_write)
        with pytest.raises(ppg.JobsFailed):
            ppg.run()
        assert Path(
            of
        ).exists()  # unlike ppg1, we know by the (lack of) output records that we need to redo this
        assert isinstance(
            ppg.global_pipegraph.last_run_result[of].error.args[0], ValueError
        )

    def test_filegenerating_ok_change_fail_ok(self):
        ppg.new(run_mode=ppg.RunMode.NOTEBOOK)
        of = "A"
        func1 = lambda of: counter("a") and write(of, "A")  # noqa: E731
        job = ppg.FileGeneratingJob(of, func1)
        ppg.run()
        assert read("a") == "1"
        ppg.run()
        assert read("a") == "1"
        assert read("A") == "A"

        def do(of):
            counter("a")
            write(of, "B")
            raise ValueError()

        job = ppg.FileGeneratingJob(of, do)
        with pytest.raises(ppg.JobsFailed):
            ppg.run()
        assert read("a") == "2"
        assert read("A") == "B"
        with pytest.raises(ppg.JobsFailed):
            ppg.run()
        assert read("a") == "3"
        assert read("A") == "B"

        ppg.FileGeneratingJob(of, func1)  # so we get the input we had previously!
        ppg.run()
        assert read("a") == "4"
        assert read("A") == "A"
        ppg.run()
        assert read("a") == "4"
        assert read("A") == "A"

    def test_simple_filegeneration_captures_stdout_stderr(self):
        of = "out/a"
        data_to_write = "hello"

        def do_write(of):
            op = open(of, "w")
            op.write(data_to_write)
            op.close()
            print("stdout is cool")
            sys.stderr.write("I am stderr")

        job = ppg.FileGeneratingJob(of, do_write)
        ppg.run()
        assert Path(of).exists()
        op = open(of, "r")
        data = op.read()
        op.close()
        assert data == data_to_write
        assert job.stdout == "stdout is cool\n"
        assert job.stderr == "I am stderr"  # no \n here

    def test_simple_filegeneration_captures_stdout_stderr_of_spawned_processe(self):
        of = "out/a"
        data_to_write = "hello"

        def do_write(of):
            import subprocess

            subprocess.check_call("echo stdout is cool", shell=True)
            subprocess.check_call("echo I am stderr>&2", shell=True)
            of.write_text("hello")

        job = ppg.FileGeneratingJob(of, do_write)
        ppg.run()
        assert Path(of).exists()
        op = open(of, "r")
        data = op.read()
        op.close()
        assert data == data_to_write
        assert job.stdout == "stdout is cool\n"
        assert job.stderr == "I am stderr\n"  # no \n here

    def test_simple_filegeneration_disabled_stdout_capture(self):
        of = "out/a"
        data_to_write = "hello"

        def do_write(of):
            write(of, data_to_write)
            print("stdout is cool")
            sys.stderr.write("I am stderr")

        job = ppg.FileGeneratingJob(of, do_write, always_capture_output=False)
        ppg.run()
        assert Path(of).exists()
        op = open(of, "r")
        data = op.read()
        op.close()
        assert data == data_to_write
        assert job.stdout == "not captured"
        assert job.stderr == "not captured"  # no \n here

    def test_simple_filegeneration_captures_stdout_stderr_giant_response(self, capsys):
        of = "out/a"

        def do_write(of):
            write(of, "hello")
            print("s" * (16 * 1024**2))
            sys.stderr.write("I" * (256 * 1024**2))

        job = ppg.FileGeneratingJob(of, do_write)
        with capsys.disabled():
            ppg.run()
        assert Path(of).exists()
        assert read(of) == "hello"
        assert job.stdout == "s" * (16 * 1024**2) + "\n"
        assert job.stderr == "I" * (256 * 1024**2)  # no \n here

    def test_simple_filegeneration_captures_stdout_stderr_failure(self):
        of = "out/a"
        data_to_write = "hello"

        def do_write(of):
            op = open(of, "w")
            op.write(data_to_write)
            op.close()
            print("stdout is cool")
            sys.stderr.write("I am stderr")
            raise ValueError()

        job = ppg.FileGeneratingJob(of, do_write)
        with pytest.raises(ppg.JobsFailed):
            ppg.run()
        assert Path(of).exists()
        op = open(of, "r")
        data = op.read()
        op.close()
        assert data == data_to_write
        assert job.stdout == "stdout is cool\n"
        assert job.stderr == "I am stderr"  # no \n here

    def test_filegeneration_does_not_change_mcp(self):
        global global_test
        global_test = 1
        of = "out/a"
        data_to_write = "hello"

        def do_write(of):
            write(of, data_to_write)
            global global_test
            global_test = 2

        ppg.FileGeneratingJob(of, do_write)
        ppg.run()
        assert global_test == 1

    def test_file_generation_chaining_simple(self):
        ofA = "out/a"

        def writeA(of):
            write(ofA, "Hello")

        jobA = ppg.FileGeneratingJob(ofA, writeA)
        ofB = "out/b"

        def writeB(of):
            op = open(ofB, "w")
            ip = open(ofA, "r")
            op.write(ip.read()[::-1])
            op.close()
            ip.close()

        jobB = ppg.FileGeneratingJob(ofB, writeB)
        jobB.depends_on(jobA)
        ppg.run()
        assert read(ofA) == read(ofB)[::-1]

    def test_file_generation_multicore(self):
        # one fork per FileGeneratingJob...
        import os

        ofA = "out/a"

        def writeA(of):
            write(ofA, "%i" % os.getpid())

        ofB = "out/b"

        def writeB(of):
            write(ofB, "%i" % os.getpid())

        ppg.FileGeneratingJob(ofA, writeA)
        ppg.FileGeneratingJob(ofB, writeB)
        ppg.run()
        assert read(ofA) != read(ofB)

    def test_filegenerating_two_jobs_same_file(self):
        ppg.new(run_mode=ppg.RunMode.NOTEBOOK)
        ppg.MultiFileGeneratingJob(
            ["out/A", "out/B"], lambda of: write("out/A", "hello")
        )
        with pytest.raises(ppg.JobOutputConflict):
            ppg.MultiFileGeneratingJob(
                ["out/A", "out/C"], lambda of: write("out/A", "world")
            )
        ppg.MultiFileGeneratingJob(
            ["out/C"], lambda of: write("out/A", "world")
        )  # that's ok, we are replacing out/C
        ppg.FileGeneratingJob(
            "out/C", lambda of: write("out/C", "C")
        )  # that's ok, we are replacing out/C (at least in relaxed /notebook mode)
        with pytest.raises(ValueError):  # just checking the inheritance
            ppg.MultiFileGeneratingJob(
                ["out/C", "out/D"], lambda of: write("out/A", "world")
            )  # note that this one does not replace out/C
        ppg.MultiFileGeneratingJob(
            ["out/D", "out/E"], lambda of: write("out/A", "world")
        )
        with pytest.raises(ppg.JobOutputConflict):
            ppg.FileGeneratingJob("out/D", lambda of: write("out/C", "C"))

    def test_multi_file_with_exing_files_rerun_to_capture_hashes(self):
        def callback(filenames):
            counter("counter")
            for f in filenames:
                f.write_text("hello")

        ppg.MultiFileGeneratingJob(["a", "b"], callback)
        Path("a").write_text("shu")
        Path("b").write_text("shu")
        ppg.run()
        assert read("a") == "hello"
        assert read("b") == "hello"
        assert read("counter") == "1"
        ppg.run()
        assert read("counter") == "1"

    def test_invaliding_removes_file(self):
        of = "out/a"
        sentinel = "out/b"

        def do_write(of):
            if Path(sentinel).exists():
                raise ValueError("second run")
            write(of, "shu")
            write(sentinel, "done")

        job = ppg.FileGeneratingJob(of, do_write, empty_ok=True)
        dep = ppg.ParameterInvariant(
            "my_params",
            {
                1,
            },
        )
        job.depends_on(dep)
        ppg.run()
        assert Path(of).exists()
        assert Path(sentinel).exists()

        ppg.new()
        job = ppg.FileGeneratingJob(of, do_write)
        dep = ppg.ParameterInvariant(
            "my_params", (2,)
        )  # same name ,changed params, job needs to rerun, but explodes...
        job.depends_on(dep)  # on average, half the mistakes are in the tests...
        with pytest.raises(ppg.JobsFailed):
            ppg.run()
        assert not (Path(of).exists())

    def test_passing_non_function(self):
        with pytest.raises(TypeError):
            ppg.FileGeneratingJob("out/a", "shu")

    def test_passing_non_string_as_jobid(self):
        with pytest.raises(TypeError):
            ppg.FileGeneratingJob(5, lambda of: 1)

    def test_exceptions_are_preserved(self):
        def shu(of):
            write("out/A", "A")
            write("out/Ay", "ax")
            raise IndexError("twenty-five")  # just some exception

        jobA = ppg.FileGeneratingJob("out/A", shu)

        with pytest.raises(ppg.JobsFailed):
            ppg.run()

        assert isinstance(
            ppg.global_pipegraph.last_run_result[jobA.job_id].error.args[0], IndexError
        )
        assert Path("out/A").exists()  #
        assert read("out/Ay") == "ax"  # but the job did run, right?

    def test_exceptions_are_preserved2(self):
        def shu(of):
            write("out/A", "A")
            write("out/Ay", "ax")
            raise TypeError("twenty-five")  # just some exception

        jobA = ppg.FileGeneratingJob("out/A", shu)

        with pytest.raises(ppg.JobsFailed):
            ppg.run()

        print(ppg.global_pipegraph.last_run_result[jobA.job_id].error)
        assert isinstance(
            ppg.global_pipegraph.last_run_result[jobA.job_id].error.args[0], TypeError
        )
        assert "twenty-five" in str(
            ppg.global_pipegraph.last_run_result[jobA.job_id].error
        )
        assert Path(
            "out/A"
        ).exists()  # should clobber the resulting files in this case - just a double check to test_invaliding_removes_file
        assert read("out/Ay") == "ax"  # but the job did run, right?


@pytest.mark.usefixtures("create_out_dir")
@pytest.mark.usefixtures("ppg2_per_test")
class TestMultiFileGeneratingJob:
    def test_basic(self):
        of = ["out/a", "out/b"]

        def do_write(of):
            for f in of:
                append(f, "shu")

        ppg.MultiFileGeneratingJob(of, do_write)
        ppg.run()
        for f in of:
            assert read(f) == "shu"
        ppg.new()
        ppg.MultiFileGeneratingJob(of, do_write)
        ppg.run()
        for f in of:
            assert read(f) == "shu"  # ie. job has net been rerun...
        # but if I now delete one...
        Path(of[0]).unlink()
        ppg.new()
        ppg.MultiFileGeneratingJob(of, do_write)
        ppg.run()
        assert read(of[0]) == "shu"
        assert (
            read(of[1]) == "shu"
        )  # Since that file was also deleted when MultiFileGeneratingJob was invalidated...

    def test_empty_raises(self):
        of = ["out/a", "out/b"]

        def do_write(of):
            for f in of:
                append(f, "shu")
            write(f, "")  # one is empty

        job = ppg.MultiFileGeneratingJob(of, do_write, empty_ok=False)
        with pytest.raises(ppg.JobsFailed):
            ppg.run()
        assert isinstance(
            ppg.global_pipegraph.last_run_result[job.job_id].error.args[0],
            ppg.JobContractError,
        )

    def test_empty_ok(self):
        of = ["out/a", "out/b"]

        def do_write(of):
            for f in of:
                append(f, "shu")
            write(f, "")  # one is empty

        ppg.MultiFileGeneratingJob(of, do_write, empty_ok=True)
        ppg.run()
        for f in of[:-1]:
            assert read(f) == "shu"
        assert read(of[-1]) == ""

    def test_exception_destroys_no_files(self):
        of = ["out/a", "out/b"]

        def do_write(of):
            for f in of:
                append(f, "shu")
            raise ValueError("explode")

        ppg.MultiFileGeneratingJob(of, do_write)
        with pytest.raises(ppg.JobsFailed):
            ppg.run()
        for f in of:
            assert Path(f).exists()

    def test_invalidation_removes_all_files(self):
        of = ["out/a", "out/b"]
        sentinel = "out/sentinel"  # hack so this one does something different the second time around...

        def do_write(of):
            if Path(sentinel).exists():
                raise ValueError("explode")
            write(sentinel, "shu")
            for f in of:
                append(f, "shu")

        ppg.MultiFileGeneratingJob(of, do_write).depends_on(
            ppg.ParameterInvariant("myparam", (1,))
        )
        ppg.run()
        for f in of:
            assert Path(f).exists()
        ppg.new()
        ppg.MultiFileGeneratingJob(of, do_write).depends_on(
            ppg.ParameterInvariant("myparam", (2,))
        )
        with pytest.raises(ppg.JobsFailed):
            ppg.run()  # since this should blow up
        for f in of:
            assert not (Path(f).exists())

    def test_passing_not_a_list_of_str(self):
        with pytest.raises(TypeError):
            ppg.MultiFileGeneratingJob("out/a", lambda of: 1)

    def test_passing_non_function(self):
        with pytest.raises(TypeError):
            ppg.MultiFileGeneratingJob(["out/a"], "shu")

    def test_exceptions_are_preserved(self):
        def shu(of):
            write("out/A", "A")
            write("out/B", "B")
            write("out/Az", "ax")
            raise IndexError("twenty-five")  # just some exception

        jobA = ppg.MultiFileGeneratingJob(["out/A", "out/B"], shu)

        with pytest.raises(ppg.JobsFailed):
            ppg.run()

        assert isinstance(
            ppg.global_pipegraph.last_run_result[jobA.job_id].error.args[0], IndexError
        )
        assert Path("out/A").exists()  # ppg1 difference
        assert Path("out/B").exists()  # ppg1 difference
        assert read("out/Az") == "ax"  # but the job did run, right?

    def raises_on_non_string_filnames(self):
        with pytest.raises(ValueError):
            ppg.MultiFileGeneratingJob(["one", 2], lambda of: write("out/A"))

    def test_raises_on_collision(self):
        with pytest.raises(ValueError):
            ppg.MultiFileGeneratingJob(["test1", "test2"], lambda of: 5)
            ppg.MultiFileGeneratingJob(["test2", "test3"], lambda of: 5)

    def test_duplicate_prevention(self):
        param = "A"
        ppg.FileGeneratingJob("out/A", lambda of: write("out/A", param))

        with pytest.raises(ValueError):
            ppg.MultiFileGeneratingJob(["out/A"], lambda of: write("out/A", param))

        ppg.new(run_mode=ppg.RunMode.NOTEBOOK)
        ppg.FileGeneratingJob("out/A", lambda of: write("out/A", param))
        ppg.MultiFileGeneratingJob(["out/A"], lambda of: write("out/A", param))

    def test_non_str(self):
        param = "A"

        with pytest.raises(TypeError):
            ppg.MultiFileGeneratingJob([25], lambda of: write("out/A", param))

    def test_non_iterable(self):
        param = "A"
        try:
            ppg.MultiFileGeneratingJob(25, lambda of: write("out/A", param))
            assert not ("Exception not raised")
        except TypeError as e:
            print(e)
            assert "files was not iterable" in str(e)

    def test_single_stre(self):
        param = "A"

        with pytest.raises(TypeError):
            ppg.MultiFileGeneratingJob("A", lambda of: write("out/A", param))


test_modifies_shared_global = []
shared_value = ""


@pytest.mark.usefixtures("create_out_dir")
@pytest.mark.usefixtures("ppg2_per_test")
class TestDataLoadingJob:
    def test_modifies_worker(self):
        # global shared
        # shared = "I was the the global in the mcp"
        def load():
            test_modifies_shared_global.append("shared data")

        of = "out/a"

        def do_write(of):
            write(
                of, "\n".join(test_modifies_shared_global)
            )  # this might actually be a problem when defining this?

        dlJo = ppg.DataLoadingJob("myjob", load)
        writejob = ppg.FileGeneratingJob(of, do_write)
        writejob.depends_on(dlJo)

        writejob2 = ppg.FileGeneratingJob(
            "out/b",
            lambda of: write("out/b", "b" + "\n".join(test_modifies_shared_global)),
        )
        writejob2.depends_on(dlJo)
        ppg.run()
        assert read(of) == "shared data"
        assert read("out/b") == "bshared data"

    def test_global_statement_works(self):
        # this currently does not work in the cloudpickle transmitted jobs -
        # two jobs refereing to global have different globals afterwards
        # or the 'global shared' does not work as expected after loading
        global shared_value
        shared_value = "I was the the global in the mcp"

        def load():
            global shared_value
            shared_value = "shared data"

        of = "out/a"

        def do_write(of):
            write(of, shared_value)

        dlJo = ppg.DataLoadingJob("myjob", load)
        writejob = ppg.FileGeneratingJob(of, do_write)
        writejob.depends_on(dlJo)
        ppg.run()
        assert read(of) == "shared data"

    def test_does_not_get_run_without_dep_job(self):
        of = "out/shu"

        def load():
            counter(of)

        ppg.DataLoadingJob("myjob", load)
        ppg.run()
        assert not Path(of).exists()
        ppg.run()
        assert not Path(of).exists()
        ppg.new()
        ppg.DataLoadingJob("myjob", load)
        ppg.run()
        assert not Path(of).exists()

    def test_does_not_get_run_in_chain_without_final_dep(self):
        of = "out/shu"

        def load():
            counter(of)

        job = ppg.DataLoadingJob("myjob", load)
        ofB = "out/sha"

        def loadB():
            counter(ofB)

        ppg.DataLoadingJob("myjobB", loadB).depends_on(job)
        ppg.run()
        assert not Path(of).exists()
        assert not Path(ofB).exists()
        ppg.run()
        assert not Path(of).exists()
        assert not Path(ofB).exists()

    def test_does_get_run_in_chain_all(self):
        of = "out/shu"

        def load():
            write(
                of, "shu"
            )  # not the fine english way, but we need a sideeffect that's checkable

        job = ppg.DataLoadingJob("myjob", load)
        ofB = "out/sha"

        def loadB():
            write(ofB, "sha")

        jobB = ppg.DataLoadingJob("myjobB", loadB).depends_on(job)
        ofC = "out/c"

        def do_write(of):
            write(ofC, ofC)

        ppg.FileGeneratingJob(ofC, do_write).depends_on(jobB)
        ppg.run()
        assert Path(of).exists()
        assert Path(ofB).exists()
        assert Path(ofC).exists()

    def test_chain_with_filegenerating_works(self):
        of = "out/a"

        def do_write(of):
            write(of, str(of))

        jobA = ppg.FileGeneratingJob(of, do_write)
        o = Dummy()

        def do_load():
            o.a = read(of)

        jobB = ppg.DataLoadingJob("loadme", do_load).depends_on(jobA)
        ofB = "out/b"

        def write2(of):
            write(ofB, o.a)

        ppg.FileGeneratingJob(ofB, write2).depends_on(jobB)
        ppg.run()
        assert read(of) == of
        assert read(ofB) == of

    def test_does_get_run_depending_on_jobgenjob(self):
        of = "out/shu"

        def load():
            write(
                of, "shu"
            )  # not the fine english way, but we need a sideeffect that's checkable

        job = ppg.DataLoadingJob("myjob", load)

        def gen():
            ofB = "out/b"

            def do_write(of):
                write(ofB, "hello")

            ppg.FileGeneratingJob(ofB, do_write)

        ppg.JobGeneratingJob("mygen", gen).depends_on(job)
        ppg.run()
        assert Path(of).exists()  # so the data loading job was run
        assert read("out/b") == "hello"  # and so was the jobgen and filegen job.

    def test_passing_non_function(self):
        with pytest.raises(TypeError):
            ppg.DataLoadingJob("out/a", "shu")

    def test_passing_non_string_as_jobid(self):
        with pytest.raises(TypeError):
            ppg.DataLoadingJob(5, lambda: 1)

    def test_failing_dataloading_jobs(self):
        o = Dummy()
        of = "out/A"

        def write(of):
            write(of, o.a)

        def load():
            o.a = "shu"
            raise ValueError()

        job_fg = ppg.FileGeneratingJob(of, write)
        job_dl = ppg.DataLoadingJob("doload", load)
        job_fg.depends_on(job_dl)
        with pytest.raises(ppg.JobsFailed):
            ppg.run()
        assert not (Path(of).exists())
        assert isinstance(
            ppg.global_pipegraph.last_run_result[job_dl.job_id].error.args[0],
            ValueError,
        )

    def test_prev_dataloading_jobs_not_done_if_there_is_a_non_dataloading_job_inbetween_that_is_done(
        self,
    ):
        # so, A = DataLoadingJob, B = FileGeneratingJob, C = DataLoadingJob, D = FileGeneratingJob
        # D.depends_on(C)
        # C.depends_on(B)
        # B.depends_on(A)
        # B is done.
        # D is not
        # since a would be loaded, and then cleaned up right away (because B is Done)
        # it should not be loaded again
        o = Dummy()

        def a():
            o.a = "A"
            append("out/A", "A")

        def b(of):
            append("out/B", "B")

        def c():
            o.c = "C"
            append("out/C", "C")

        def d(of):
            append("out/D", "D")

        jobA = ppg.DataLoadingJob("out/A", a, depend_on_function=False)
        jobB = ppg.FileGeneratingJob("out/B", b, depend_on_function=False)
        jobB.depends_on(jobA)
        ppg.run()
        assert read("out/A") == "A"
        assert read("out/B") == "B"

        jobC = ppg.DataLoadingJob("out/C", c, depend_on_function=False)
        jobD = ppg.FileGeneratingJob("out/D", d, depend_on_function=False)
        jobD.depends_on(jobC)
        jobC.depends_on(jobB)
        ppg.run()
        assert read("out/A") == "A"
        assert read("out/B") == "B"
        assert read("out/C") == "C"
        assert read("out/D") == "D"

    def test_sending_a_non_pickable_exception_data_loading(self):
        class UnpickableException(Exception):
            def __getstate__(self):
                raise ValueError("Can't pickle me")

        def load():
            raise UnpickableException()

        jobA = ppg.DataLoadingJob("out/A", load)
        jobB = ppg.FileGeneratingJob("out/B", lambda of: True)
        jobB.depends_on(jobA)

        with pytest.raises(ppg.JobsFailed):
            ppg.run()

        assert isinstance(
            ppg.global_pipegraph.last_run_result[jobA.job_id].error.args[0],
            UnpickableException,
        )

    def test_sending_a_non_pickable_exception_file_generating(self):
        class UnpickableException(Exception):
            def __getstate__(self):
                raise ValueError("Can't pickle me")

        def load(of):
            raise UnpickableException()

        jobB = ppg.FileGeneratingJob("out/B", load)

        with pytest.raises(ppg.JobsFailed):
            ppg.run()
        assert isinstance(
            ppg.global_pipegraph.last_run_result[jobB.job_id].error.args[0],
            ppg.exceptions.JobDied,
        )

        assert isinstance(
            ppg.global_pipegraph.last_run_result[jobB.job_id].error.args[0].args[0], str
        )

        assert (
            "UnpickableException"
            in ppg.global_pipegraph.last_run_result[jobB.job_id].error.args[0].args[0]
        )

    def test_creating_jobs_in_file_generating_are_ignored(self):
        def load(of):
            ppg.global_pipegraph.new_jobs = (
                {}
            )  # just to see if we can reach the check in the resource coordinator!
            c = ppg.FileGeneratingJob("out/C", lambda of: write("out/C", "C"))
            write("out/A", "A")
            return [c]

        ppg.FileGeneratingJob("out/A", load)
        ppg.run()
        assert read("out/A") == "A"
        assert not Path("out/C").exists()

    def test_creating_jobs_in_data_loading(self):
        def load():
            from loguru import logger  # noqa:F401

            logger.info("in load")
            ppg.FileGeneratingJob(
                "out/C", lambda of: write("out/C", "C"), depend_on_function=False
            )

        a = ppg.FileGeneratingJob(
            "out/A", lambda of: write("out/A", "A"), depend_on_function=False
        )
        b = ppg.DataLoadingJob("out/B", load, depend_on_function=False)
        a.depends_on(b)
        ppg.run()
        assert Path("out/C").exists()
        # it is ok to create jobs in dataloading .
        # ppg1 guarded against this, but it had to special case everything
        # around on-the-fly-jobs
        # the drawback here is that the DataLoadingJob might not run,
        # but perhaps that's just what you want.

    def test_accept_path_as_job_id(self):
        ppg.DataLoadingJob(Path("shu"), lambda: 55)

    def test_job_gen_does_not_clobber_history_of_input_jobs(self):
        a = ppg.FileGeneratingJob("a", lambda of: counter("A") and of.write_text("a"))
        b = ppg.JobGeneratingJob(
            "b",
            lambda: counter("B")
            and ppg.FileGeneratingJob(
                "c", lambda of: counter("C") and of.write_text("c")
            ),
        )
        b.depends_on(a)
        ppg.run()
        assert Path("a").read_text() == "a"
        assert Path("A").read_text() == "1"
        assert Path("B").read_text() == "1"
        assert Path("C").read_text() == "1"
        ppg.run()
        assert Path("a").read_text() == "a"
        assert Path("A").read_text() == "1"
        assert Path("B").read_text() == "2"
        assert Path("C").read_text() == "1"

    def test_dict_return(self):
        collector = []

        def gen():
            collector.append("a")
            return {"hello": 123, "world": "world"}

        a = ppg.DataLoadingJob("gen", gen)
        b = ppg.FileGeneratingJob("b", lambda of: of.write_text(collector[0]))
        b.depends_on(a)
        ppg.run()
        assert read("b") == "a"


@pytest.mark.usefixtures("create_out_dir")
@pytest.mark.usefixtures("ppg2_per_test")
class TestAttributeJob:
    def test_basic_attribute_loading(self):
        o = Dummy()

        def load():
            return "shu"

        job = ppg.AttributeLoadingJob("load_dummy_shu", o, "a", load)
        of = "out/a"

        def do_write(of):
            write(of, o.a)

        ppg.FileGeneratingJob(of, do_write).depends_on(job)
        ppg.run()
        assert read(of) == "shu"

    def test_chained(self):
        o = Dummy()

        def load():
            return "shu"

        def load2():
            return o.a + "sha"

        job = ppg.AttributeLoadingJob("load_dummy_shu", o, "a", load)
        job2 = ppg.AttributeLoadingJob("load_dummy_shu2", o, "b", load2)
        of = "out/a"

        def do_write(of):
            write(of, o.b)

        ppg.FileGeneratingJob(of, do_write).depends_on(job2)
        job2.depends_on(job)
        ppg.run()
        assert read(of) == "shusha"

    def test_attribute_loading_does_not_affect_mcp(self):
        o = Dummy()

        def load():
            return "shu"

        job = ppg.AttributeLoadingJob("load_dummy_shu", o, "a", load)
        of = "out/a"

        def do_write(of):
            write(of, o.a)

        ppg.FileGeneratingJob(of, do_write).depends_on(job)
        ppg.run()
        assert read(of) == "shu"
        assert not (hasattr(o, "a"))

    def test_attribute_loading_does_not_run_without_dependency(self):
        o = Dummy()
        tf = "out/testfile"

        def load():
            counter(tf)
            return "shu"

        ppg.AttributeLoadingJob(
            "load_dummy_shu", o, "a", load, depend_on_function=False
        )
        ppg.run()
        assert not (hasattr(o, "a"))  # never assigned
        assert not Path("tf").exists()
        ppg.run()
        assert not Path("tf").exists()
        assert not (hasattr(o, "a"))

    def test_attribute_loading_does_run_without_dependency_if_invalidated(self):
        o = Dummy()
        tf = "out/testfile"

        def load():
            write(tf, "hello")
            return "shu"

        ppg.AttributeLoadingJob("load_dummy_shu", o, "a", load)
        ppg.run()
        assert not Path(tf).exists()
        assert not (hasattr(o, "a"))

    def test_attribute_disappears_after_direct_dependency(self):
        o = Dummy()
        job = ppg.AttributeLoadingJob("load_dummy_shu", o, "a", lambda: "shu")
        of = "out/A"

        def do_write(of):
            write(of, o.a)

        fgjob = ppg.FileGeneratingJob(of, do_write).depends_on(job)
        of2 = "out/B"

        def later_write(of2):
            raise ValueError()
            write(of2, o.a)

        # might be pure luck that this job runs after the cleanup
        ppg.FileGeneratingJob(of2, later_write).depends_on(fgjob)
        with pytest.raises(ppg.JobsFailed):
            ppg.run()
        assert read(of) == "shu"
        assert not (Path(of2).exists())

    def ppg1_test_attribute_disappears_after_direct_dependencies(self):
        """I can't get tihs test to run in ppg2 - the cleanup does happen,
        but this can't show it. It is now a job, so
        it 's not a given that it runs before B or C
        (actually, I believe test_attribute_disappears_after_direct_dependency
        only works by sheer accident as well?)

        """
        o = Dummy()
        job = ppg.AttributeLoadingJob("load_dummy_shu", o, "a", lambda: "shu")
        of = "out/A"

        def do_write(of):
            write(of, o.a)

        fgjob = ppg.FileGeneratingJob(of, do_write).depends_on(job)
        of2 = "out/B"

        def later_write(of2):
            write(of2, o.a)

        fgjobB = ppg.FileGeneratingJob(of2, later_write).depends_on(
            fgjob
        )  # now, this one does not depend on job, o it should not be able to access o.a
        of3 = "out/C"

        def also_write(of3):
            write(of3, o.a)

        fgjobC = ppg.FileGeneratingJob(of3, also_write).depends_on(job)
        fgjobB.depends_on(
            fgjobC
        )  # otherwise, B might be started C returned, and the cleanup will not have occurred!
        with pytest.raises(ppg.JobsFailed):
            ppg.run()
            pass
        assert read(of) == "shu"
        assert read(of3) == "shu"
        assert not (Path(of2).exists())

    def test_passing_non_string_as_attribute(self):
        o = Dummy()

        with pytest.raises(ValueError):
            ppg.AttributeLoadingJob("out/a", o, 5, 55)

    def test_passing_non_function(self):
        o = Dummy()

        with pytest.raises(TypeError):
            ppg.AttributeLoadingJob("out/a", o, "a", 55)

    def test_passing_non_string_as_jobid(self):
        o = Dummy()

        with pytest.raises(TypeError):
            ppg.AttributeLoadingJob(5, o, "a", lambda: 55)

    def test_no_swapping_attributes_for_one_job(self):
        def cache():
            return list(range(0, 100))

        o = Dummy()
        ppg.AttributeLoadingJob("out/A", o, "a", cache)

        with pytest.raises(ppg.JobRedefinitionError):
            ppg.AttributeLoadingJob("out/A", o, "b", cache)

    def test_raises_on_non_string_attribute_name(self):
        with pytest.raises(ValueError):
            o = Dummy()
            ppg.AttributeLoadingJob("out/A", o, 23, lambda: 5)

    def test_raises_on_non_function_callback(self):
        with pytest.raises(ValueError):
            o = Dummy()
            ppg.AttributeLoadingJob("out/A", o, 23, 55)

    def test_no_swapping_objects_for_one_job(self):
        def cache():
            return list(range(0, 100))

        o = Dummy()
        o2 = Dummy()
        ppg.CachedAttributeLoadingJob("out/A", o, "a", cache)

        with pytest.raises(ppg.JobRedefinitionError):
            ppg.CachedAttributeLoadingJob("out/A", o2, "a", cache)

    def test_no_swapping_callbacks(self):

        o = Dummy()
        ppg.AttributeLoadingJob("out/A", o, "a", lambda: 55, depend_on_function=False)

        with pytest.raises(ppg.JobRedefinitionError):
            ppg.AttributeLoadingJob(
                "out/A", o, "a", lambda: 55 + 1, depend_on_function=False
            )

    def test_no_swapping_callbacks_cached(self):

        o = Dummy()
        ppg.CachedAttributeLoadingJob(
            "out/A", o, "a", lambda: 55, depend_on_function=False
        )

        with pytest.raises(ppg.JobRedefinitionError):
            ppg.CachedAttributeLoadingJob(
                "out/A", o, "a", lambda: 55 + 1, depend_on_function=False
            )

    def test_ignore_code_changes(self):
        def a():
            append("out/Aa", "A")
            return "5"

        o = Dummy()
        jobA = ppg.AttributeLoadingJob("out/A", o, "a", a, depend_on_function=False)
        jobB = ppg.FileGeneratingJob("out/B", lambda of: write("out/B", o.a))
        jobB.depends_on(jobA)
        ppg.run()
        assert read("out/Aa") == "A"
        assert read("out/B") == "5"
        ppg.new()

        def b():
            append("out/Aa", "B")
            return "5"

        jobA = ppg.AttributeLoadingJob("out/A", o, "a", b, depend_on_function=False)
        jobB = ppg.FileGeneratingJob("out/B", lambda of: write("out/B", o.a))
        jobB.depends_on(jobA)
        ppg.run()
        # not rerun
        assert read("out/Aa") == "A"
        assert read("out/B") == "5"

    def test_callback_must_be_callable(self):
        with pytest.raises(TypeError):
            o = Dummy()
            ppg.AttributeLoadingJob("load_dummy_shu", o, "a", "shu")

    def test_returning_none(self):
        ppg.new(run_mode=ppg.RunMode.NOTEBOOK)
        o = Dummy()
        a = ppg.FileGeneratingJob("out/a", lambda of: counter("a") and write(of, "A"))
        b = ppg.FileGeneratingJob("out/b", lambda of: counter("b") and write(of, "B"))
        c = ppg.AttributeLoadingJob("o", o, "o", lambda: counter("c") and None)
        b.depends_on(a)
        a.depends_on_params("x")
        a.depends_on(c)
        b.depends_on(c)
        ppg.run()
        assert read("a") == "1"
        assert read("b") == "1"
        assert read("c") == "1"
        a.depends_on_params("y")
        ppg.run()
        assert read("a") == "2"
        assert read("c") == "2"
        assert read("b") == "1"  # ppg2_rust


@pytest.mark.usefixtures("create_out_dir")
@pytest.mark.usefixtures("ppg2_per_test")
class TestTempFileGeneratingJob:
    def test_basic(self):
        temp_file = "out/temp"

        def write_temp(of):
            write(temp_file, "hello")

        temp_job = ppg.TempFileGeneratingJob(temp_file, write_temp)
        ofA = "out/A"

        def write_A(of):
            write(ofA, read(temp_file))

        fgjob = ppg.FileGeneratingJob(ofA, write_A)
        fgjob.depends_on(temp_job)
        ppg.run()
        assert read(ofA) == "hello"
        assert not (Path(temp_file).exists())

    def test_does_not_get_return_if_output_is_done(self):
        temp_file = "out/temp"
        out_file = "out/A"
        count_file = "out/count"
        normal_count_file = "out/countA"

        def write_count(of):
            try:
                count = read(out_file)
                count = count[: count.find(":")]
            except IOError:
                count = "0"
            count = int(count) + 1
            write(out_file, str(count) + ":" + read(temp_file))
            append(normal_count_file, "A")

        def write_temp(of):
            write(temp_file, "temp")
            append(count_file, "X")

        jobA = ppg.FileGeneratingJob(out_file, write_count)
        jobTemp = ppg.TempFileGeneratingJob(temp_file, write_temp)
        jobA.depends_on(jobTemp)
        ppg.run()
        assert not (Path(temp_file).exists())
        assert read(out_file) == "1:temp"
        assert read(count_file) == "X"
        assert read(normal_count_file) == "A"
        # now, rerun. Tempfile has been deleted,
        # and should not be regenerated
        ppg.new()
        jobA = ppg.FileGeneratingJob(out_file, write_count)
        jobTemp = ppg.TempFileGeneratingJob(temp_file, write_temp)
        jobA.depends_on(jobTemp)
        ppg.run()
        assert not (Path(temp_file).exists())
        assert read(out_file) == "1:temp"
        assert read(count_file) == "X"
        assert read(normal_count_file) == "A"

    def test_does_not_get_return_if_output_is_not(self):
        temp_file = "out/temp"
        out_file = "out/A"
        count_file = "out/count"
        normal_count_file = "out/countA"

        def write_count(out_file):
            try:
                count = read(out_file)
                count = count[: count.find(":")]
            except IOError:
                count = "0"
            count = int(count) + 1
            write(out_file, str(count) + ":" + read(temp_file))
            append(normal_count_file, "A")

        def write_temp(of):
            write(temp_file, "temp")
            append(count_file, "X")

        jobA = ppg.FileGeneratingJob(out_file, write_count)
        jobTemp = ppg.TempFileGeneratingJob(temp_file, write_temp)
        jobA.depends_on(jobTemp)
        ppg.run()
        assert not (Path(temp_file).exists())
        assert read(out_file) == "1:temp"
        assert read(count_file) == "X"
        assert read(normal_count_file) == "A"
        # now, rerun. Tempfile has been deleted,
        # and should  be regenerated
        Path(out_file).unlink()
        ppg.new()
        jobA = ppg.FileGeneratingJob(out_file, write_count)
        jobTemp = ppg.TempFileGeneratingJob(temp_file, write_temp)
        jobA.depends_on(jobTemp)
        ppg.run()
        assert read(out_file) == "1:temp"  # since the outfile was removed...
        assert read(count_file) == "XX"
        assert read(normal_count_file) == "AA"
        assert not (Path(temp_file).exists())

    def test_dependand_explodes(self, job_trace_log):
        temp_file = "out/temp"

        def write_temp(of):
            append(of, "hello")

        temp_job = ppg.TempFileGeneratingJob(temp_file, write_temp)
        ofA = "out/A"

        def write_A(of):
            raise ValueError("shu")

        fgjob = ppg.FileGeneratingJob(ofA, write_A)
        fgjob.depends_on(temp_job)
        with pytest.raises(ppg.JobsFailed):
            ppg.run()
        # ppg.run()
        assert not (Path(ofA).exists())
        assert Path(temp_file).exists()

        ppg.new()

        def write_A_ok(ofA):
            write(ofA, read(temp_file))

        temp_job = ppg.TempFileGeneratingJob(temp_file, write_temp)
        fgjob = ppg.FileGeneratingJob(ofA, write_A_ok)
        fgjob.depends_on(temp_job)
        ppg.run()

        assert read(ofA) == "hello"  # tempfile job has not been rerun
        assert not (Path(temp_file).exists())  # and the tempfile has been removed...

    def test_removes_tempfile_on_exception(self):
        temp_file = "out/temp"

        def write_temp(of):
            write(temp_file, "hello")
            raise ValueError("should")

        temp_job = ppg.TempFileGeneratingJob(temp_file, write_temp)
        ofA = "out/A"

        def write_A(of):
            write(ofA, read(temp_file))

        fgjob = ppg.FileGeneratingJob(ofA, write_A)
        fgjob.depends_on(temp_job)
        with pytest.raises(ppg.JobsFailed):
            ppg.run()
        assert Path(temp_file).exists()
        assert not (Path(ofA).exists())

    def test_passing_non_function(self):
        with pytest.raises(TypeError):
            ppg.TempFileGeneratingJob("out/a", "shu")

    def test_passing_non_string_as_jobid(self):
        with pytest.raises(TypeError):
            ppg.TempFileGeneratingJob(5, lambda of: 1)

    def test_rerun_because_of_new_dependency_does_not_rerun_old(self):
        jobA = ppg.FileGeneratingJob(
            "out/A",
            lambda of: append("out/A", read("out/temp")) or append("out/Ab", "A"),
        )
        jobB = ppg.TempFileGeneratingJob("out/temp", lambda of: write("out/temp", "T"))
        jobA.depends_on(jobB)
        ppg.run()
        assert not (Path("out/temp").exists())
        assert read("out/A") == "T"
        assert read("out/Ab") == "A"  # ran once

        ppg.new()
        jobA = ppg.FileGeneratingJob(
            "out/A", lambda of: append("out/A", read("out/temp"))
        )
        jobB = ppg.TempFileGeneratingJob("out/temp", lambda of: write("out/temp", "T"))
        jobA.depends_on(jobB)
        jobC = ppg.FileGeneratingJob(
            "out/C", lambda of: append("out/C", read("out/temp"))
        )
        jobC.depends_on(jobB)
        ppg.run()
        assert not (Path("out/temp").exists())
        assert read("out/Ab") == "A"  # ran once, not rewritten
        assert read("out/C") == "T"  # a new file

    def test_chaining_multiple(self):
        jobA = ppg.TempFileGeneratingJob("out/A", lambda of: write("out/A", "A"))
        jobB = ppg.TempFileGeneratingJob(
            "out/B", lambda of: write("out/B", read("out/A") + "B")
        )
        jobC = ppg.TempFileGeneratingJob(
            "out/C", lambda of: write("out/C", read("out/A") + "C")
        )
        jobD = ppg.FileGeneratingJob(
            "out/D", lambda of: write("out/D", read("out/B") + read("out/C"))
        )
        jobD.depends_on(jobC)
        jobD.depends_on(jobB)
        jobC.depends_on(jobA)
        jobB.depends_on(jobA)
        ppg.run()
        assert read("out/D") == "ABAC"
        assert not (Path("out/A").exists())
        assert not (Path("out/B").exists())
        assert not (Path("out/C").exists())

    def test_chaining_multiple_differently(self):
        jobA = ppg.TempFileGeneratingJob("out/A", lambda of: write("out/A", "A"))
        jobB = ppg.TempFileGeneratingJob(
            "out/B", lambda of: write("out/B", read("out/A") + "B")
        )
        jobD = ppg.FileGeneratingJob(
            "out/D", lambda of: write("out/D", read("out/B") + "D")
        )
        jobE = ppg.FileGeneratingJob(
            "out/E", lambda of: write("out/E", read("out/B") + "E")
        )
        jobF = ppg.FileGeneratingJob(
            "out/F", lambda of: write("out/F", read("out/A") + "F")
        )
        jobD.depends_on(jobB)
        jobE.depends_on(jobB)
        jobB.depends_on(jobA)
        jobF.depends_on(jobA)
        ppg.run()
        assert read("out/D") == "ABD"
        assert read("out/E") == "ABE"
        assert read("out/F") == "AF"
        assert not (Path("out/A").exists())
        assert not (Path("out/B").exists())
        assert not (Path("out/C").exists())

    def test_rerun_because_of_new_dependency_does_not_rerun_old_chained(self):
        jobA = ppg.TempFileGeneratingJob("out/A", lambda of: write("out/A", "A"))
        jobB = ppg.TempFileGeneratingJob(
            "out/B", lambda of: write("out/B", read("out/A") + "B")
        )
        jobC = ppg.FileGeneratingJob(
            "out/C",
            lambda of: write("out/C", read("out/B") + "C") or append("out/Cx", "1"),
        )
        jobB.depends_on(jobA)
        jobC.depends_on(jobB)
        ppg.run()
        assert read("out/C") == "ABC"
        assert read("out/Cx") == "1"

        ppg.new()
        jobA = ppg.TempFileGeneratingJob("out/A", lambda of: write("out/A", "A"))
        jobB = ppg.TempFileGeneratingJob(
            "out/B", lambda of: write("out/B", read("out/A") + "B")
        )
        jobC = ppg.FileGeneratingJob(
            "out/C",
            lambda of: write("out/C", read("out/B") + "C") or append("out/Cx", "1"),
        )
        jobD = ppg.FileGeneratingJob(
            "out/D",
            lambda of: write("out/D", read("out/A") + "D") or append("out/Dx", "1"),
        )
        jobB.depends_on(jobA)
        jobC.depends_on(jobB)
        jobD.depends_on(jobA)
        ppg.run()
        assert read("out/D") == "AD"
        assert read("out/Dx") == "1"
        assert read("out/C") == "ABC"
        assert read("out/Cx") == "1"

        ppg.new()
        jobA = ppg.TempFileGeneratingJob(
            "out/A", lambda of: write("out/A", "a")
        )  # note changing function code!
        jobB = ppg.TempFileGeneratingJob(
            "out/B", lambda of: write("out/B", read("out/A") + "B")
        )
        jobC = ppg.FileGeneratingJob(
            "out/C",
            lambda of: write("out/C", read("out/B") + "C") or append("out/Cx", "1"),
        )
        jobD = ppg.FileGeneratingJob(
            "out/D",
            lambda of: write("out/D", read("out/A") + "D") or append("out/Dx", "1"),
        )
        jobB.depends_on(jobA)
        jobC.depends_on(jobB)
        jobD.depends_on(jobA)
        ppg.run()
        assert read("out/D") == "aD"
        assert read("out/Dx") == "11"  # both get rerun
        assert read("out/C") == "aBC"
        assert read("out/Cx") == "11"

    def test_cleanup_if_never_run(self):
        temp_file = "out/temp"

        def write_temp(of):
            write(temp_file, "hello")

        def write_a(of):
            write("A", "hello")

        temp_job = ppg.TempFileGeneratingJob(
            temp_file, write_temp, depend_on_function=False
        )
        ppg.FileGeneratingJob("A", write_a, depend_on_function=False)
        write_a("A")  # so the file is there!
        ppg.run()
        assert not (Path("out/temp").exists())
        ppg.new()
        write_temp(temp_file)
        assert Path("out/temp").exists()
        # this job never runs...
        ppg.TempFileGeneratingJob(temp_file, write_temp, depend_on_function=False)
        # temp_job.do_cleanup_if_was_never_run = True
        ppg.run()
        assert Path("out/temp").exists()  # no run, no cleanup

    def test_cleanup_if_fails(self, job_trace_log):
        def fail(of):
            of.write_text("hello")
            raise ValueError("thisisit")

        a = ppg.TempFileGeneratingJob("a", fail)
        b = ppg.FileGeneratingJob("b", lambda of: of.write_text(read("a")))
        b.depends_on(a)
        with pytest.raises(ppg.JobsFailed):
            b()
        assert Path("a").exists()  # cleanup does not run
        assert not Path("b").exists()
        assert "thisisit" in str(a.exception)
        assert "Upstream" in str(b.exception)

    def test_temp_ds_fail_not_rerun(self, job_trace_log):
        def tf(of):
            of.write_text("A")
            counter("tf")

        jtf = ppg.TempFileGeneratingJob("A", tf)

        def j(of):
            if counter("j") == "0":
                raise ValueError()
            of.write_text("J")

        jj = ppg.FileGeneratingJob("J", j)
        jj.depends_on(jtf)
        with pytest.raises(ppg.JobsFailed):
            ppg.run()
        assert Path("A").read_text() == "A"
        assert Path("tf").read_text() == "1"
        assert Path("j").read_text() == "1"
        assert not Path("J").exists()
        ppg.run()
        assert Path("J").read_text() == "J"
        assert Path("j").read_text() == "2"
        assert Path("tf").read_text() == "1"

    def test_file_already_presen(self):
        def doit(output_filename):
            counter("tf")
            output_filename.write_text("done")

        j = ppg.TempFileGeneratingJob(".ppg/deleteme", doit)
        j2 = ppg.FileGeneratingJob(
            ".ppg/shu", lambda of: counter("j2") and of.write_text("hello")
        )
        j2.depends_on(j)
        j2()
        assert Path("j2").read_text() == "1"
        assert Path("tf").read_text() == "1"

        Path(".ppg/shu").unlink()
        Path(".ppg/deleteme").write_text("done")
        j2()
        assert Path("tf").read_text() == "1"  # same hash, file present
        assert Path("j2").read_text() == "2"

    def test_adding_removing_outputs_does_not_trigger_downstreams_that_depend_on_the_unchanged_inputs(
        self,
    ):
        def do_a(ofs):
            counter("a")
            if isinstance(ofs, Path):
                ofs = [ofs]
            for of in ofs:
                of.write_text(of.name)

        jobA = ppg.MultiFileGeneratingJob(["A", "B"], do_a)
        jobC = ppg.FileGeneratingJob(
            "C", lambda of: counter("c") and of.write_text(of.name)
        )
        jobC.depends_on("A")  # note how this depends on the file "A", not the job jobA.
        ppg.run()
        assert Path("C").read_text() == "C"
        assert Path("c").read_text() == "1"

        ppg.new()
        jobA = ppg.MultiFileGeneratingJob(["A", "B", "D"], do_a)
        jobC = ppg.FileGeneratingJob(
            "C", lambda of: counter("c") and of.write_text(of.name)
        )
        jobC.depends_on("A")
        print("part2")
        ppg.run()
        assert Path("D").read_text() == "D"
        assert Path("C").read_text() == "C"
        assert Path("a").read_text() == "2"
        assert Path("c").read_text() == "1"

        # now let's take it out again.
        print("part3")
        ppg.new()
        jobA = ppg.MultiFileGeneratingJob(["A", "B"], do_a)
        jobC = ppg.FileGeneratingJob(
            "C", lambda of: counter("c") and of.write_text(of.name)
        )
        jobC.depends_on("A")
        ppg.run()
        assert Path("D").read_text() == "D"
        assert Path("C").read_text() == "C"
        assert Path("a").read_text() == "3"
        assert Path("c").read_text() == "1"

        # take one out & put one in *at the same time*
        ppg.new()
        jobA = ppg.MultiFileGeneratingJob(["A", "B", "E"], do_a)
        jobC = ppg.FileGeneratingJob(
            "C", lambda of: counter("c") and of.write_text(of.name)
        )
        jobC.depends_on("A")
        ppg.run()
        assert Path("D").read_text() == "D"
        assert Path("C").read_text() == "C"
        assert Path("a").read_text() == "4"
        assert Path("c").read_text() == "1"

        # now turn it into a FG
        ppg.new()
        jobA = ppg.FileGeneratingJob("A", do_a)
        jobC = ppg.FileGeneratingJob(
            "C", lambda of: counter("c") and of.write_text(of.name)
        )
        jobC.depends_on("A")
        ppg.run()
        assert Path("D").read_text() == "D"
        assert Path("C").read_text() == "C"
        assert Path("a").read_text() == "5"
        assert Path("c").read_text() == "1"

        # now turn it into back into a MFG
        ppg.new()
        ppg.MultiFileGeneratingJob(["A", "B"], do_a)
        jobC = ppg.FileGeneratingJob(
            "C", lambda of: counter("c") and of.write_text(of.name)
        )
        jobC.depends_on("A")
        ppg.run()
        assert Path("D").read_text() == "D"
        assert Path("C").read_text() == "C"
        assert Path("a").read_text() == "6"
        assert Path("c").read_text() == "1"


@pytest.mark.usefixtures("create_out_dir")
@pytest.mark.usefixtures("ppg2_per_test")
class TestMultiTempFileGeneratingJob:
    def test_basic(self):
        temp_files = ["out/temp", "out/temp2"]

        def write_temp(of):
            for temp_file in temp_files:
                write(temp_file, "hello")

        temp_job = ppg.MultiTempFileGeneratingJob(temp_files, write_temp)
        ofA = "out/A"

        def write_A(of):
            write(ofA, read(temp_files[0]) + read(temp_files[1]))

        fgjob = ppg.FileGeneratingJob(ofA, write_A)
        fgjob.depends_on(temp_job)
        ppg.run()
        assert read(ofA) == "hellohello"
        assert not (Path(temp_files[0]).exists())
        assert not (Path(temp_files[1]).exists())

    def test_basic_dependes_were_done(self):
        temp_files = ["out/temp", "out/temp2"]

        def write_temp(of):
            write("temp_sentinel", "one")
            for temp_file in temp_files:
                write(temp_file, "hello")

        temp_job = ppg.MultiTempFileGeneratingJob(
            temp_files, write_temp, depend_on_function=False
        )
        ofA = "out/A"

        def write_A(of):
            write(ofA, read(temp_files[0]) + read(temp_files[1]))

        fgjob = ppg.FileGeneratingJob(ofA, write_A, depend_on_function=False)
        fgjob.depends_on(temp_job)
        write(ofA, "two")
        ppg.run()
        assert (
            read(ofA) == "hellohello"
        )  # change from ppg1 - we rerun if we don't have a hash recorded
        assert Path("temp_sentinel").exists()

    def raises_on_non_string_filnames(self):
        with pytest.raises(ValueError):
            ppg.MultiTempFileGeneratingJob(["one", 2], lambda of: write("out/A"))

    def test_raises_on_collision(self):
        with pytest.raises(ppg.JobOutputConflict):
            ppg.MultiTempFileGeneratingJob(["test1", "test2"], lambda of: 5)
            ppg.MultiTempFileGeneratingJob(["test2", "test3"], lambda of: 5)

    def test_duplicate_prevention(self):
        param = "A"
        ppg.FileGeneratingJob("out/A", lambda of: write("out/A", param))

        with pytest.raises(ppg.JobRedefinitionError):
            ppg.MultiTempFileGeneratingJob(["out/A"], lambda of: write("out/A", param))

    def test_non_str(self):
        param = "A"

        with pytest.raises(TypeError):
            ppg.MultiTempFileGeneratingJob([25], lambda of: write("out/A", param))

    def test_non_iterable(self):
        param = "A"
        with pytest.raises(TypeError):
            ppg.MultiTempFileGeneratingJob(25, lambda of: write("out/A", param))


@pytest.mark.usefixtures("ppg2_per_test")
class TestNoDotDotInJobIds:
    def test_no_dot_dot(self):
        """all ../ must be resolved before it becomes a job id"""
        from unittest.mock import patch

        collector = set()
        org_dedup = ppg.jobs._dedup_job

        def collecting_dedup(cls, job_id):
            collector.add(job_id)
            return org_dedup(cls, job_id)

        with patch("pypipegraph2.jobs._dedup_job", collecting_dedup):
            j = ppg.MultiFileGeneratingJob(["something/../shu"], lambda of: 5)
            assert j.job_id in collector
            assert not ".." in j.job_id
            assert not "something/../shu" in collector
            collector.clear()

            j = ppg.FileGeneratingJob("something/../shu2", lambda of: 5)
            assert j.job_id in collector
            assert not ".." in j.job_id
            assert not "something/../shu2" in collector
            collector.clear()

            j = ppg.FileInvariant("something/../shu3")
            assert j.job_id in collector
            assert not ".." in j.job_id
            assert not "something/../shu3" in collector
            collector.clear()

            with pytest.raises(TypeError):  # we don't resolve function invariant names
                ppg.FunctionInvariant("something/../shu3b")

            j = ppg.TempFileGeneratingJob("something/../shu4", lambda of: 5)
            assert j.job_id in collector
            assert not ".." in j.job_id
            assert not "something/../shu4" in collector
            collector.clear()

            j = ppg.MultiTempFileGeneratingJob(["something/../shu5"], lambda of: 5)
            assert j.job_id in collector
            assert not ".." in j.job_id
            assert not "something/../shu4" in collector
            collector.clear()

            o = object()
            j = ppg.CachedAttributeLoadingJob("something/../shu6", o, "attr", lambda: 5)
            assert j.calc.job_id in collector
            assert j.load.job_id in collector
            assert not ".." in j.calc.job_id
            assert not ".." in j.load.job_id
            assert not "something/../shu6" in collector
            assert not "loadsomething/../shu6" in collector
            collector.clear()

            j = ppg.CachedDataLoadingJob(
                "something/../shu7", lambda: 5, lambda value: 5
            )
            assert j.calc.job_id in collector
            assert j.load.job_id in collector
            assert not ".." in j.calc.job_id
            assert not ".." in j.load.job_id
            assert not "something/../shu7" in collector
            collector.clear()

            j = ppg.PlotJob(
                "something/or_other/../shu.png", lambda: None, lambda data: None
            )
            assert j.plot.job_id in collector
            assert j.cache.calc.job_id in collector
            assert j.cache.load.job_id in collector
            assert j.table.job_id in collector
            for job_id in collector:
                assert not ".." in job_id



import os
import gzip
from pathlib import Path
from loguru import logger
import stat
import time
import hashlib
import shutil
import pytest
import pypipegraph2 as ppg
from .shared import write, read, append, Dummy, counter, force_load


class Undepickable(object):
    def __getstate__(self):
        return {"shu": 123}  # must not return falsey value

    def __setstate__(self, state):
        self.sha = state["shu"]
        import pickle

        raise pickle.UnpicklingError("SHU")


@pytest.mark.usefixtures("create_out_dir")
@pytest.mark.usefixtures("ppg2_per_test")
class TestInvariant:
    def sentinel_count(self):
        sentinel = "out/sentinel"
        try:
            op = open(sentinel, "r")
            count = int(op.read())
            op.close()
        except IOError:
            count = 1
        op = open(sentinel, "w")
        op.write("%i" % (count + 1))
        op.close()
        return count

    def test_filegen_jobs_detect_code_change(self):
        of = "out/a"

        def do_write(of):
            append(of, "shu" * self.sentinel_count())

        ppg.FileGeneratingJob(of, do_write)
        ppg.run()
        assert read(of) == "shu"
        ppg.new()
        ppg.FileGeneratingJob(of, do_write)
        ppg.run()
        assert read(of) == "shu"  # has not been run again...

        def do_write2(of):
            append(of, "sha")

        ppg.new()
        ppg.FileGeneratingJob(of, do_write2)
        ppg.run()
        assert read(of) == "sha"  # has been run again ;).

    def test_filegen_jobs_ignores_code_change(self):
        of = "out/a"

        def do_write(of):
            counter("A")
            append(of, "shu" * self.sentinel_count())

        ppg.FileGeneratingJob(of, do_write)
        ppg.run()

        assert read(of) == "shu"
        assert read("A") == "1"
        ppg.new()
        ppg.FileGeneratingJob(of, do_write)
        ppg.run()
        assert read(of) == "shu"  # has not been run again, for no change
        assert read("A") == "1"

        ppg.new()

        def do_write2(of):
            counter("A")
            append(of, "sha")

        ppg.FileGeneratingJob(of, do_write2, depend_on_function=False)
        ppg.run()
        assert read(of) == "sha"  # has been run again - number of invariants changed!
        assert read("A") == "2"

        ppg.new()
        ppg.FileGeneratingJob(of, do_write2)
        ppg.run()
        assert read(of) == "sha"  # Readding the invariant does trigger again
        assert read("A") == "3"

    def test_parameter_dependency(self):
        of = "out/a"

        def do_write(of):
            append(of, "shu" * self.sentinel_count())

        job = ppg.FileGeneratingJob(of, do_write)
        param_dep = ppg.ParameterInvariant("myparam", (1, 2, 3))
        job.depends_on(param_dep)
        ppg.run()
        assert read(of) == "shu"
        ppg.new()
        job = ppg.FileGeneratingJob(of, do_write)
        param_dep = ppg.ParameterInvariant("myparam", (1, 2, 3))
        job.depends_on(param_dep)
        ppg.run()
        assert read(of) == "shu"  # has not been run again...
        ppg.new()
        job = ppg.FileGeneratingJob(of, do_write)
        param_dep = ppg.ParameterInvariant("myparam", (1, 2, 3, 4))
        job.depends_on(param_dep)
        ppg.run()
        assert read(of) == "shushu"  # has been run again ;).

    def test_parameter_invariant_adds_hidden_job_id_prefix(self):
        param = "A"
        jobA = ppg.FileGeneratingJob("out/A", lambda of: write("out/A", param))
        jobB = ppg.ParameterInvariant("out/A", param)
        jobA.depends_on(jobB)
        ppg.run()
        assert read("out/A") == param

    def test_depends_on_func(self):
        a = ppg.FileGeneratingJob("out/A", lambda of: write("a"))
        b, a_again = a.depends_on_func("a123", lambda: 123)
        assert b.job_id.startswith("FI" + a.job_id + "_")
        assert ppg.global_pipegraph.has_edge(b, a)
        assert a_again is a

    def test_depends_on_file(self):
        a = ppg.FileGeneratingJob("out/A", lambda of: write("a"))
        write("shu", "hello")
        b = a.depends_on_file("shu")
        assert b.self is a
        assert ppg.global_pipegraph.has_edge(b.invariant, a)

    def test_depends_on_params(self):
        a = ppg.FileGeneratingJob("out/A", lambda of: write("a"))
        b = a.depends_on_params(23)
        assert b.invariant.job_id == "PIout/A"
        # assert b.invariant.parameters == 23
        assert ppg.global_pipegraph.has_edge(b.invariant, a)
        assert b.self is a

    def test_parameter_invariant_twice_different_values(self):
        ppg.ParameterInvariant("a", (1, 2, 3))
        with pytest.raises(ValueError):
            ppg.ParameterInvariant("a", (1, 2, 4))

    def test_filetime_dependency(self):
        of = "out/a"

        def do_write(of):
            append(of, "shu" * self.sentinel_count())

        ftfn = "out/ftdep"
        write(ftfn, "hello")
        write(of, "hello")
        job = ppg.FileGeneratingJob(of, do_write)
        dep = ppg.FileInvariant(ftfn)
        job.depends_on(dep)
        ppg.run()
        assert (
            read(of) == "shu"
        )  # job get's run though there is a file, because the FileInvariant was not stored before...
        ppg.new()
        job = ppg.FileGeneratingJob(of, do_write)
        dep = ppg.FileInvariant(ftfn)
        job.depends_on(dep)
        ppg.run()
        assert read(of) == "shu"  # job does not get rerun...

        time.sleep(1)  # so linux actually advances the file time in the next line
        write(ftfn, "hello")  # same content, different time

        ppg.new()
        job = ppg.FileGeneratingJob(of, do_write)
        dep = ppg.FileInvariant(ftfn)
        job.depends_on(dep)
        ppg.run()
        assert (
            read(of) == "shu"
        )  # job does not get rerun - filetime invariant is now filechecksum invariant...

    def test_file_did_not_exist(self):
        ppg.FileInvariant("shu")
        with pytest.raises(ppg.JobsFailed):
            ppg.run()
        assert "did not exist" in str(ppg.global_pipegraph.last_run_result["shu"].error)

    def test_filechecksum_dependency_raises_on_too_short_a_filename(self):
        ppg.global_pipegraph.allow_short_filenames = False
        with pytest.raises(ValueError):
            ppg.FileInvariant("a")

        with pytest.raises(ValueError):
            ppg.FileInvariant("sh")
        ppg.FileInvariant("shu")

    def test_filechecksum_dependency(self):
        of = "out/a"

        def do_write(of):
            append(of, "shu" * self.sentinel_count())

        ftfn = "out/ftdep"
        write(ftfn, "hello")
        # import stat
        # logging.info('file time after creating %s'% os.stat(ftfn)[stat.ST_MTIME])

        write(of, "hello")

        job = ppg.FileGeneratingJob(of, do_write)
        dep = ppg.FileInvariant(ftfn)
        job.depends_on(dep)
        ppg.run()
        assert (
            read(of) == "shu"
        )  # job get's run though there is a file, because the FileInvariant was not stored before...
        ppg.new()
        job = ppg.FileGeneratingJob(of, do_write)
        dep = ppg.FileInvariant(ftfn)
        job.depends_on(dep)
        ppg.run()
        assert read(of) == "shu"  # job does not get rerun...

        time.sleep(1)  # so linux actually advances the file time in the next line
        # logging.info("NOW REWRITE")
        write(ftfn, "hello")  # same content, different time

        ppg.new()
        job = ppg.FileGeneratingJob(of, do_write)
        dep = ppg.FileInvariant(ftfn)
        job.depends_on(dep)
        ppg.run()
        assert read(of) == "shu"  # job does not get rerun...

        # time.sleep(1) #we don't care about the time, size should be enough...
        write(ftfn, "hello world!!")  # different time
        time.sleep(1)  # give the file system a second to realize the change.

        ppg.new()
        job = ppg.FileGeneratingJob(of, do_write)
        dep = ppg.FileInvariant(ftfn)
        job.depends_on(dep)
        ppg.run()
        assert read(of) == "shushu"  # job does get rerun

    @pytest.mark.skip  # TODO: Renaming support?
    def test_input_file_was_renamed(self):
        of = "out/B"

        def do_write(of):
            append(of, "shu" * self.sentinel_count())

        ftfn = "out/ftdep"
        write(ftfn, "hello")
        # import stat
        # logging.info('file time after creating %s'% os.stat(ftfn)[stat.ST_MTIME])

        write(of, "hello")

        job = ppg.FileGeneratingJob(of, do_write)
        dep = ppg.FileInvariant(ftfn)
        job.depends_on(dep)
        ppg.run()
        assert (
            read(of) == "shu"
        )  # job get's run though there is a file, because the FileInvariant was not stored before...

        ppg.new()
        job = ppg.FileGeneratingJob(of, do_write)
        dep = ppg.FileInvariant(ftfn)
        job.depends_on(dep)
        ppg.run()
        assert read(of) == "shu"  # job does not get rerun...

        os.mkdir("out/moved_here")
        shutil.move(ftfn, os.path.join("out/moved_here", "ftdep"))
        ppg.new()
        job = ppg.FileGeneratingJob(of, do_write)
        dep = ppg.FileInvariant(os.path.join("out/moved_here", "ftdep"))
        job.depends_on(dep)
        assert read(of) == "shu"  # job does not get rerun...
        ppg.run()
        assert read(of) == "shu"  # job does not get rerun...

    @pytest.mark.skip()  # I have no idea why this was useful. Possibly the PrebuildJobs support?
    def test_file_invariant_with_md5sum(self):
        of = "out/a"

        def do_write(of):
            append(of, "shu" * self.sentinel_count())

        ftfn = "out/ftdep"
        write(ftfn, "hello")
        # import stat
        # logging.info('file time after creating %s'% os.stat(ftfn)[stat.ST_MTIME])

        write(of, "hello")

        job = ppg.FileGeneratingJob(of, do_write)
        dep = ppg.FileInvariant(ftfn)
        job.depends_on(dep)
        ppg.run()
        assert (
            read(of) == "shu"
        )  # job get's run though there is a file, because the FileInvariant was not stored before...

        with open(ftfn + ".md5sum", "wb") as op:
            op.write(hashlib.md5(b"hello world").hexdigest().encode("utf-8"))
        write(ftfn, "hello world")  # different content
        t = time.time()
        # now make
        os.utime(ftfn, (t, t))
        os.utime(ftfn + ".md5sum", (t, t))
        time.sleep(1)  # give the file system a second to realize the change.

        ppg.new()
        job = ppg.FileGeneratingJob(of, do_write)
        dep = ppg.FileInvariant(ftfn)
        job.depends_on(dep)
        ppg.run()
        assert (
            read(of) == "shushu"
        )  # job get's run though there is a file, because the md5sum changed.

        with open(ftfn + ".md5sum", "wb") as op:
            op.write(hashlib.md5(b"hello world").hexdigest().encode("utf-8"))
        write(ftfn, "hello")  # different content, but the md5sum is stil the same!
        t = time.time()
        # now make
        os.utime(ftfn, (t, t))
        os.utime(ftfn + ".md5sum", (t, t))
        time.sleep(1)  # give the file system a second to realize the change.

        ppg.new()
        job = ppg.FileGeneratingJob(of, do_write)
        dep = ppg.FileInvariant(ftfn)
        job.depends_on(dep)
        ppg.run()
        assert read(of) == "shushu"  # job does not get rerun, md5sum did not change...

        t = time.time() - 100  # force a file time mismatch
        os.utime(
            ftfn, (t, t)
        )  # I must change the one on the actual file, otherwise the 'size+filetime is the same' optimization bytes me

        ppg.new()
        job = ppg.FileGeneratingJob(of, do_write)
        dep = ppg.FileInvariant(ftfn)
        job.depends_on(dep)
        ppg.run()
        assert (
            read(of) == "shushushu"
        )  # job does get rerun, md5sum and file time mismatch
        assert os.stat(ftfn)[stat.ST_MTIME] == os.stat(ftfn + ".md5sum")[stat.ST_MTIME]

    def test_invariant_dumping_on_job_failure(self):
        def w(of):
            write("out/A", "A")
            append("out/B", "B")

        def func_c(of):
            append("out/C", "C")

        func_dep = ppg.FunctionInvariant("func_c", func_c)
        fg = ppg.FileGeneratingJob("out/A", w, depend_on_function=False)
        fg.depends_on(func_dep)
        ppg.run()
        assert read("out/A") == "A"
        assert read("out/B") == "B"
        ppg.new()

        def func_c1(of):
            append("out/C", "D")

        def w2(of):
            raise ValueError()  # so there is an error in a job...

        func_dep = ppg.FunctionInvariant("func_c", func_c1)  # so this invariant changes
        fg = ppg.FileGeneratingJob(
            "out/A", w2, depend_on_function=False
        )  # and this job crashes
        fg.depends_on(func_dep)
        with pytest.raises(ppg.JobsFailed):
            ppg.run()
        assert not (os.path.exists("out/A"))  # since it was removed, and not recreated
        assert read("out/B") == "B"
        ppg.new()
        func_dep = ppg.FunctionInvariant(
            "func_c", func_c1
        )  # no invariant change this time
        fg = ppg.FileGeneratingJob(
            "out/A", w, depend_on_function=False
        )  # but this was not done the last time...
        fg.depends_on(func_dep)
        ppg.run()
        assert read("out/A") == "A"
        assert read("out/B") == "BB"

    def test_invariant_dumping_on_graph_exception(self, mocker):
        # when an exception occurs not within a job
        # but within the pipegraph itself (e.g. when the user hit's CTRL-C
        # during history dumping
        # which we simulate here
        def w(of):
            write("out/A", "A")
            append("out/B", "B")

        def func_c(of):
            append("out/C", "C")

        func_dep = ppg.FunctionInvariant("func_c", func_c)
        fg = ppg.FileGeneratingJob("out/A", w, depend_on_function=False)
        fg.depends_on(func_dep)
        ppg.run()
        assert read("out/A") == "A"
        assert read("out/B") == "B"

        ppg.new(run_mode=ppg.RunMode.CONSOLE)
        # ppg.new()

        def func_c1(of):
            append("out/C", "D")

        def w2(of):
            write("out/A", "A2")
            raise ValueError()  # so there is an error in a job...

        func_dep = ppg.FunctionInvariant("func_c", func_c1)  # so this invariant changes
        fg = ppg.FileGeneratingJob(
            "out/A", w2, depend_on_function=False
        )  # and this job crashes
        fg.depends_on(func_dep)  # so a get's deleted, and rebuild
        ppg.FileGeneratingJob(
            "out/C", lambda of: counter("out/c") and append(of, "C")
        )
        assert not hasattr(ppg.global_pipegraph, "last_run_result")
        ppg.global_pipegraph._test_failing_outside_of_job = True
        with pytest.raises(ppg.JobsFailed):
            ppg.run()
        print(ppg.global_pipegraph.do_raise)
        assert isinstance(ppg.global_pipegraph.do_raise[0], KeyboardInterrupt)
        assert len(ppg.global_pipegraph.do_raise) == 2
        assert hasattr(ppg.global_pipegraph, "last_run_result")
        assert os.path.exists("out/A")  # The file get's written.
        assert read("out/B") == "B"
        assert read("out/C") == "C"
        assert read("out/c") == "1"  #
        mocker.stopall()

        ppg.new()
        func_dep = ppg.FunctionInvariant(
            "func_c", func_c1
        )  # no invariant change this time
        # but we had no stored input/output for A, right?
        # so it get's rerun
        fg = ppg.FileGeneratingJob(
            "out/A", w, depend_on_function=False
        )  # but this was not done the last time...
        fg.depends_on(func_dep)
        ppg.FileGeneratingJob(
            "out/C", lambda of: counter("out/c") and append(of, "C")
        )
        ppg.run()
        assert read("out/A") == "A"
        assert read("out/B") == "BB"
        assert read("out/C") == "C"  #
        assert read("out/c") == "1"  # c did not get rerun

    def test_sig_int_is_ignored_in_console_mode(self):
        ppg.new(run_mode=ppg.RunMode.CONSOLE)

        def sigint():
            import signal

            counter("a")
            os.kill(os.getpid(), signal.SIGINT)
            counter("A")

        job = ppg.DataLoadingJob("A", sigint)
        force_load(job)
        ppg.run()
        assert read("a") == "1"
        assert read("A") == "1"

    def test_input_output_dumping_dies_for_some_reason(self, ppg2_per_test, mocker):
        ppg.global_pipegraph._test_failing_outside_of_job = True
        ppg.FileGeneratingJob("A", lambda of: counter("a") and write(of, "A"))
        ppg.FileGeneratingJob("B", lambda of: counter("b") and write(of, "B"))
        with pytest.raises(ppg.RunFailedInternally):
            ppg.run()
        assert read("A") == "A"
        assert read("B") == "B"
        assert read("a") == "1"
        assert read("b") == "1"
        ppg.run()
        assert read("A") == "A"
        assert read("B") == "B"
        assert (
            read("a") == "1"
        )  # we had captured the output for both. history saving now is all or nothing...
        assert read("b") == "1"

    def test_FileInvariant_cant_have_dependencies(self):
        # invariants are always roots of the DAG - they can't have any dependencies themselves
        write("out/shu", "shu")
        job = ppg.FileInvariant("out/shu")
        jobB = ppg.FileGeneratingJob("out/A", lambda of: write("out/A", "a"))

        with pytest.raises(ppg.JobContractError):
            job.depends_on(jobB)

    def test_FunctionInvariant_cant_have_dependencies(self):
        # invariants are always roots of the DAG - they can't have any dependencies themselves
        job = ppg.FunctionInvariant("shu", lambda: 55)
        jobB = ppg.FileGeneratingJob("out/A", lambda of: write("out/A", "a"))

        with pytest.raises(ppg.JobContractError):
            job.depends_on(jobB)

    def test_ParameterInvariant_cant_have_dependencies(self):
        # invariants are always roots of the DAG - they can't have any dependencies themselves
        job = ppg.ParameterInvariant("out/shu", ("123",))
        jobB = ppg.FileGeneratingJob("out/A", lambda of: write("out/A", "a"))

        with pytest.raises(ppg.JobContractError):
            job.depends_on(jobB)

    def test_invariant_loading_issues_on_value_catastrophic(self):
        a = ppg.DataLoadingJob("a", lambda: 5)
        b = ppg.FileGeneratingJob(
            "out/b", lambda of: write("out/b", "b"), depend_on_function=False
        )
        b.depends_on(a)
        write("out/b", "a")
        import pickle

        Path(ppg.global_pipegraph.get_history_filename()).parent.mkdir(parents=True)
        with gzip.GzipFile(ppg.global_pipegraph.get_history_filename(), "wb") as op:
            pickle.dump(a.job_id, op, pickle.HIGHEST_PROTOCOL)
            op.write(b"This breaks")
        with pytest.raises(ppg.HistoryLoadingFailed):
            ppg.run()
        assert read("out/b") == "a"  # job was not run

    def test_file_invariant_swapping(self, ppg2_per_test):
        Path("a").write_text("a")
        Path("b").write_text("b")

        def out(of):
            counter("counter")
            of.write_text(Path("a").read_text() + Path("b").read_text()),

        job = ppg.FileGeneratingJob("c", out, depend_on_function=False)
        job.depends_on_file("a")
        job.depends_on_file("b")
        ppg.run()
        assert read("c") == "ab"
        assert read("counter") == "1"
        ppg.run()
        assert read("counter") == "1"
        ppg2_per_test.new()
        job = ppg.FileGeneratingJob("c", out, depend_on_function=False)
        job.depends_on_file("b")
        job.depends_on_file("a")
        ppg.run()
        assert read("counter") == "1"

    def test_file_invariant_replaced(self):
        Path("a.tsv").write_text("a")
        a = ppg.FileInvariant("a.tsv")

        def func(of):
            counter("J")
            of.write_text("j")

        j = ppg.FileGeneratingJob("j", func)
        j.depends_on(a)
        ppg.run()
        assert read("j") == "j"
        assert read("J") == "1"
        ppg.run()
        assert read("J") == "1"
        ppg.new()
        Path("b.tsv").write_text("b")
        b = ppg.FileInvariant("b.tsv")
        j = ppg.FileGeneratingJob("j", func)
        j.depends_on(b)
        ppg.run()
        assert read("J") == "2"


def first_value(d):
    return list(d.values())[0]


@pytest.mark.usefixtures("ppg2_per_test")
class TestFunctionInvariant:
    # most of the function invariant testing is handled by other test classes.
    # but these are more specialized.

    def test_generator_expressions(self):
        def get_func(r):
            def shu():
                return sum(i + 0 for i in r)

            return shu

        def get_func2(r):
            def shu():
                return sum(i + 0 for i in r)

            return shu

        def get_func3(r):
            def shu():
                return sum(i + 1 for i in r)

            return shu

        a = ppg.FunctionInvariant("a", get_func(100))
        b = ppg.FunctionInvariant(
            "b", get_func2(100)
        )  # that invariant should be the same
        c = ppg.FunctionInvariant(
            "c", get_func3(100)
        )  # and this invariant should be different
        av = a.run(None, None)
        bv = b.run(None, None)
        cv = c.run(None, None)
        assert a.run(None, None)
        assert ppg.FunctionInvariant.compare_hashes(
            None, first_value(av), first_value(bv)
        )
        assert not ppg.FunctionInvariant.compare_hashes(
            None, first_value(av), first_value(cv)
        )

    def test_lambdas(self):
        def get_func(x):
            def inner():
                arg = lambda y: x + x + x  # noqa:E731
                return arg(1)

            return inner

        def get_func2(x):
            def inner():
                arg = lambda y: x + x + x  # noqa:E731
                return arg(1)

            return inner

        def get_func3(x):
            def inner():
                arg = lambda y: x + x  # noqa:E731
                return arg(1)

            return inner

        a = ppg.FunctionInvariant("a", get_func(100))
        b = ppg.FunctionInvariant(
            "b", get_func2(100)
        )  # that invariant should be the same
        c = ppg.FunctionInvariant(
            "c", get_func3(100)
        )  # and this invariant should be different
        av = a.run(None, None)
        bv = b.run(None, None)
        cv = c.run(None, None)
        self.maxDiff = 20000
        assert a.run(None, None)
        assert ppg.FunctionInvariant.compare_hashes(
            None, first_value(av), first_value(bv)
        )
        assert not ppg.FunctionInvariant.compare_hashes(
            None, first_value(av), first_value(cv)
        )

    def test_inner_functions(self):
        def get_func(x):
            def inner():
                return 23

            return inner

        def get_func2(x):
            def inner():
                return 23

            return inner

        def get_func3(x):
            def inner():
                return 23 + 5

            return inner

        a = ppg.FunctionInvariant("a", get_func(100))
        b = ppg.FunctionInvariant(
            "b", get_func2(100)
        )  # that invariant should be the same
        c = ppg.FunctionInvariant(
            "c", get_func3(100)
        )  # and this invariant should be different
        av = a.run(None, None)
        bv = b.run(None, None)
        cv = c.run(None, None)
        assert a.run(None, None)
        assert ppg.FunctionInvariant.compare_hashes(
            None, first_value(av), first_value(bv)
        )
        assert not ppg.FunctionInvariant.compare_hashes(
            None, first_value(av), first_value(cv)
        )

    def test_nested_inner_functions(self):
        def get_func(x):
            def inner():
                def shu():
                    return 23

                return shu

            return inner

        def get_func2(x):
            def inner():
                def shu():
                    return 23

                return shu

            return inner

        def get_func3(x):
            def inner():
                def shu():
                    return 23 + 5

                return shu

            return inner

        a = ppg.FunctionInvariant("a", get_func(100))
        b = ppg.FunctionInvariant(
            "b", get_func2(100)
        )  # that invariant should be the same
        c = ppg.FunctionInvariant(
            "c", get_func3(100)
        )  # and this invariant should be different
        av = a.run(None, None)
        bv = b.run(None, None)
        cv = c.run(None, None)
        assert a.run(None, None)
        assert ppg.FunctionInvariant.compare_hashes(
            None, first_value(av), first_value(bv)
        )
        assert not ppg.FunctionInvariant.compare_hashes(
            None, first_value(av), first_value(cv)
        )

    def test_inner_functions_with_parameters(self):
        def get_func(x):
            def inner():
                return x

            return inner

        a = ppg.FunctionInvariant("a", get_func(100))
        b = ppg.FunctionInvariant(
            "b", get_func(100)
        )  # that invariant should be the same
        c = ppg.FunctionInvariant(
            "c", get_func(2000)
        )  # and this invariant should be different
        av = a.run(None, None)
        bv = b.run(None, None)
        cv = c.run(None, None)
        assert a.run(None, None)
        assert ppg.FunctionInvariant.compare_hashes(
            None, first_value(av), first_value(bv)
        )
        assert not ppg.FunctionInvariant.compare_hashes(
            None, first_value(av), first_value(cv)
        )

    def test_passing_non_function_raises(self):
        with pytest.raises(TypeError):
            ppg.FunctionInvariant("out/a", "shu")

    def test_passing_none_as_function_is_ok(self, create_out_dir):
        job = ppg.FunctionInvariant("out/a", None)
        str(job)
        jobB = ppg.FileGeneratingJob("out/A", lambda of: write("out/A", "A"))
        jobB.depends_on(job)
        ppg.run()
        assert read("out/A") == "A"

    def test_passing_non_string_as_jobid(self):
        with pytest.raises(TypeError):
            ppg.FunctionInvariant(5, lambda: 1)

    def test_cant_have_dependencies(self):
        # invariants are always roots of the DAG - they can't have any dependencies themselves
        def shu():
            pass

        job = ppg.FunctionInvariant("shu", shu)
        jobB = ppg.FileGeneratingJob("out/A", lambda of: write("out/A", "a"))

        with pytest.raises(ppg.JobContractError):
            job.depends_on(jobB)

    def test_raises_on_duplicate_with_different_functions(self):
        def shu():
            return "a"

        ppg.FunctionInvariant("A", shu)
        ppg.FunctionInvariant("A", shu)  # ok.
        with pytest.raises(ppg.JobRedefinitionError):
            ppg.FunctionInvariant("A", lambda: "b")  # raises ValueError

        def sha():
            def shu():
                return "b"

            return shu

        ppg.FunctionInvariant("B", sha())
        ppg.FunctionInvariant("B", sha())

    def test_instance_functions_ok(self, create_out_dir):
        class shu:
            def __init__(self, letter):
                self.letter = letter

            def get_job(self):
                job = ppg.FileGeneratingJob(
                    "out/" + self.letter, lambda of: append(of, "A")
                )
                job.depends_on(ppg.FunctionInvariant("shu.sha", self.sha))
                return job

            def sha(self):
                return 55 * 23

        x = shu("A")
        x.get_job()
        ppg.run()
        assert read("out/A") == "A"
        append("out/A", "A")

        ppg.new()
        x.get_job()
        y = shu("B")
        j1 = y.get_job()
        j2 = y.get_job()
        assert ppg.FunctionInvariant.functions_equal(
            j1.generating_function, j2.generating_function
        )
        # assert j1 is j2 # todo: interactive/notebook differences

    def test_buildin_function(self):
        a = ppg.FunctionInvariant("a", open)
        assert "<built-in" in str(a)

    def test_function_invariant_non_function(self):
        class CallMe:
            def __call__(self):
                raise ValueError()

        a = ppg.FunctionInvariant("a", CallMe)
        with pytest.raises(
            ValueError
        ):  # todo: is this the right behaviour? can't we just forward to __call__ as the invariant?
            a.run(None, None)

    def test_closure_capturing(self):
        def func(da_list):
            def f():
                return da_list

            return f

        a = ppg.FunctionInvariant("a", func([1, 2, 3]))
        b = ppg.FunctionInvariant(
            "b", func([1, 2, 3])
        )  # that invariant should be the same
        c = ppg.FunctionInvariant(
            "c", func([1, 2, 3, 4])
        )  # and this invariant should be different
        av = a.run(None, None)
        bv = b.run(None, None)
        cv = c.run(None, None)
        assert a.run(None, None)
        assert ppg.FunctionInvariant.compare_hashes(
            None, first_value(av), first_value(bv)
        )
        assert not ppg.FunctionInvariant.compare_hashes(
            None, first_value(av), first_value(cv)
        )

    def test_function_to_str_builtin(self):
        assert ppg.FunctionInvariant.function_to_str(open) == "<built-in function open>"

    def test_closure_capturing_dict(self):
        def func(da_list):
            def f():
                return da_list

            return f

        a = ppg.FunctionInvariant("a", func({"1": "a", "3": "b", "2": "c"}))
        b = ppg.FunctionInvariant(
            "b", func({"1": "a", "3": "b", "2": "c"})
        )  # that invariant should be the same
        c = ppg.FunctionInvariant(
            "c", func({"1": "a", "3": "b", "2": "d"})
        )  # and this invariant should be different
        av = a.run(None, None)
        bv = b.run(None, None)
        cv = c.run(None, None)
        assert a.run(None, None)
        assert ppg.FunctionInvariant.compare_hashes(
            None, first_value(av), first_value(bv)
        )
        assert not ppg.FunctionInvariant.compare_hashes(
            None, first_value(av), first_value(cv)
        )

    def test_closure_capturing_set(self):
        def func(da_list):
            def f():
                return da_list

            return f

        import random

        x = set(["1", "2", "3", "4", "5", "6", "7", "8"])
        a = ppg.FunctionInvariant("a", func(x))
        x2 = list(x)
        random.shuffle(x2)
        x2 = set(x2)
        b = ppg.FunctionInvariant("b", func(x2))  # that invariant should be the same
        c = ppg.FunctionInvariant(
            "c", func({"3", "2"})
        )  # and this invariant should be different
        av = a.run(None, None)
        bv = b.run(None, None)
        cv = c.run(None, None)
        assert a.run(None, None)
        assert ppg.FunctionInvariant.compare_hashes(
            None, first_value(av), first_value(bv)
        )
        assert not ppg.FunctionInvariant.compare_hashes(
            None, first_value(av), first_value(cv)
        )

    def test_closure_capturing_frozen_set(self):
        def func(da_list):
            def f():
                return da_list

            return f

        import random

        x = frozenset(["1", "2", "3", "4", "5", "6", "7", "8"])
        a = ppg.FunctionInvariant("a", func(x))
        x2 = list(x)
        random.shuffle(x2)
        x2 = frozenset(x2)
        b = ppg.FunctionInvariant("b", func(x2))  # that invariant should be the same
        c = ppg.FunctionInvariant(
            "c", func(frozenset({"3", "2"}))
        )  # and this invariant should be different
        av = a.run(None, None)
        bv = b.run(None, None)
        cv = c.run(None, None)
        assert a.run(None, None)
        assert ppg.FunctionInvariant.compare_hashes(
            None, first_value(av), first_value(bv)
        )
        assert not ppg.FunctionInvariant.compare_hashes(
            None, first_value(av), first_value(cv)
        )

    def test_function_invariants_are_equal_if_dis_identical_or_source_identical(self):
        python_version = (3, 99)
        a = {"source": "hello", python_version: ("dis", "closure")}
        b = a.copy()
        b["source"] = "hello_world"
        c = a.copy()
        c[python_version] = ("disB", "closure")
        assert ppg.FunctionInvariant.compare_hashes(None, a, a, python_version)
        assert ppg.FunctionInvariant.compare_hashes(
            None, a, b, python_version
        )  # same dis ,different source
        assert not ppg.FunctionInvariant.compare_hashes(
            None, a, c, python_version
        )  # different dis, same source

    def test_source_file_mtime_change_without_hash_change(self):
        def inner():
            pass

        # python_version = tuple(sys.version_info)[:2]  # we only care about major.minor

        a = ppg.FunctionInvariant("a", inner)
        calc = a.run(None, None)
        changed = calc.copy()
        changed["FIa"]["source_file"]["mtime"] = -1
        changed["FIa"]["source_file"]["dis"] = "find me"
        calc2 = a.run(None, changed)
        assert calc2["FIa"]["source_file"]["dis"] == "find me"


@pytest.mark.usefixtures("create_out_dir")
@pytest.mark.usefixtures("ppg2_per_test")
class TestDependency:
    def test_simple_chain(self):
        o = Dummy()

        def load_a():
            return "shu"

        jobA = ppg.AttributeLoadingJob("a", o, "myattr", load_a)
        ofB = "out/B"

        def do_write_b(ofB):
            write(ofB, o.myattr)

        jobB = ppg.FileGeneratingJob(ofB, do_write_b).depends_on(jobA)
        ofC = "out/C"

        def do_write_C(ofC):
            write(ofC, o.myattr)

        ppg.FileGeneratingJob(ofC, do_write_C).depends_on(jobA)

        ofD = "out/D"

        def do_write_d(ofD):
            write(ofD, read(ofC) + read(ofB))

        ppg.FileGeneratingJob(ofD, do_write_d).depends_on([jobA, jobB])

    def test_failed_job_kills_those_after(self):
        ofA = "out/A"

        def write_a(ofA):
            append(ofA, "hello")

        jobA = ppg.FileGeneratingJob(ofA, write_a)

        ofB = "out/B"

        def write_b(ofB):
            raise ValueError("shu")

        jobB = ppg.FileGeneratingJob(ofB, write_b)
        jobB.depends_on(jobA)

        ofC = "out/C"

        def write_c(ofC):
            write(ofC, "hello")

        jobC = ppg.FileGeneratingJob(ofC, write_c)
        jobC.depends_on(jobB)
        with pytest.raises(ppg.JobsFailed):
            ppg.run()
        assert os.path.exists(ofA)  # which was before the error
        assert not (os.path.exists(ofB))  # which was on the error
        assert not (os.path.exists(ofC))  # which was after the error
        ppg.new()
        jobA = ppg.FileGeneratingJob(ofA, write_a)
        jobC = ppg.FileGeneratingJob(ofC, write_c)

        def write_b_ok(ofB):
            write(ofB, "BB")

        jobB = ppg.FileGeneratingJob(ofB, write_b_ok)
        jobB.depends_on(jobA)
        jobC.depends_on(jobB)
        ppg.run()

        assert os.path.exists(ofA)
        assert read(ofA) == "hello"  # run only once!
        assert os.path.exists(ofB)
        assert os.path.exists(ofC)

    def test_done_filejob_does_not_gum_up_execution(self):
        ofA = "out/A"
        write(ofA, "1111")

        def write_a(ofA):
            append(ofA, "hello")

        jobA = ppg.FileGeneratingJob(ofA, write_a, depend_on_function=False)

        ofB = "out/B"

        def write_b(ofB):
            append(ofB, "hello")

        jobB = ppg.FileGeneratingJob(ofB, write_b)
        jobB.depends_on(jobA)

        ofC = "out/C"

        def write_c(ofC):
            write(ofC, "hello")

        jobC = ppg.FileGeneratingJob(ofC, write_c)
        jobC.depends_on(jobB)
        assert os.path.exists(ofA)

        ppg.run()

        assert os.path.exists(ofB)
        assert os.path.exists(ofC)
        assert (
            read(ofA) == "hello"
        )  # change from ppgA, if it's not our file (we have recorded not output), rerun

    def test_invariant_violation_redoes_deps_but_not_nondeps(self):
        def get_job(name):
            fn = "out/" + name

            def do_write(of):
                if os.path.exists(fn + ".sentinel"):
                    d = read(fn + ".sentinel")
                else:
                    d = ""
                append(fn + ".sentinel", name)  # get's longer all the time...
                write(fn, d + name)  # get's deleted anyhow...

            return ppg.FileGeneratingJob(fn, do_write)

        jobA = get_job("A")
        jobB = get_job("B")
        jobC = get_job("C")
        get_job("D")
        jobC.depends_on(jobB)
        jobB.depends_on(jobA)
        dep = ppg.ParameterInvariant("myparam", ("hello",))
        jobA.depends_on(dep)
        ppg.run()
        assert read("out/A") == "A"
        assert read("out/B") == "B"
        assert read("out/C") == "C"

        ppg.new()
        jobA = get_job("A")
        jobB = get_job("B")
        jobC = get_job("C")
        get_job("D")
        jobC.depends_on(jobB)
        jobB.depends_on(jobA)
        dep = ppg.ParameterInvariant("myparam", ("hello stranger",))
        jobA.depends_on(dep)  # now, the invariant has been changed, all jobs rerun...
        ppg.run()
        assert read("out/A") == "AA"  # thanks to our smart rerun aware job definition..
        assert read("out/B") == "BB"
        assert read("out/C") == "CC"
        assert read("out/D") == "D"  # since that one does not to be rerun...

    def test_depends_on_accepts_a_list(self):
        jobA = ppg.FileGeneratingJob("out/A", lambda of: write("out/A", "A"))
        jobB = ppg.FileGeneratingJob("out/B", lambda of: write("out/B", "B"))
        jobC = ppg.FileGeneratingJob("out/C", lambda of: write("out/C", "C"))
        jobC.depends_on([jobA, jobB])
        ppg.run()
        assert read("out/A") == "A"
        assert read("out/B") == "B"
        assert read("out/C") == "C"

    def test_job_iter(self):
        jobA = ppg.FileGeneratingJob("out/A", lambda of: write("out/A", "A"))
        ll = list(iter(jobA))
        assert ll[0] is jobA

    def test_depends_on_accepts_multiple_values(self):
        jobA = ppg.FileGeneratingJob("out/A", lambda of: write("out/A", "A"))
        jobB = ppg.FileGeneratingJob("out/B", lambda of: write("out/B", "B"))
        jobC = ppg.FileGeneratingJob("out/C", lambda of: write("out/C", "C"))
        jobC.depends_on(jobA, jobB)
        ppg.run()
        assert read("out/A") == "A"
        assert read("out/B") == "B"
        assert read("out/C") == "C"

    def test_depends_on_accepts_multiple_values_mixed(self):
        jobA = ppg.FileGeneratingJob("out/A", lambda of: write("out/A", "A"))
        jobB = ppg.FileGeneratingJob("out/B", lambda of: write("out/B", "B"))
        jobC = ppg.FileGeneratingJob("out/C", lambda of: write("out/C", "C"))
        jobC.depends_on(jobA, [jobB])
        ppg.run()
        assert read("out/A") == "A"
        assert read("out/B") == "B"
        assert read("out/C") == "C"

    def test_depends_on_none_ignored(self):
        jobA = ppg.FileGeneratingJob("out/A", lambda of: write("out/A", "A"))
        jobB = ppg.FileGeneratingJob("out/B", lambda of: write("out/B", "B"))
        jobC = ppg.FileGeneratingJob("out/C", lambda of: write("out/C", "C"))
        jobC.depends_on(jobA, [jobB], None, [None])
        jobC.depends_on(None)
        jobC.depends_on()  # that's a no-op as well
        jobC.depends_on([])  # that's a no-op as well
        ppg.run()
        assert read("out/A") == "A"
        assert read("out/B") == "B"
        assert read("out/C") == "C"

    def test_depends_on_excludes_on_non_jobs(self):
        jobA = ppg.FileGeneratingJob("out/A", lambda of: write("out/A", "A"))

        with pytest.raises(KeyError):
            jobA.depends_on("SHU")

    def test_depends_on_instant_cycle_check(self):
        jobA = ppg.FileGeneratingJob("out/A", lambda of: write("out/A", "A"))
        jobB = ppg.FileGeneratingJob("out/b", lambda of: write("out/B", "b"))
        jobB.depends_on(jobA)

        with pytest.raises(ppg.NotADag):
            jobA.depends_on(jobA)

        with pytest.raises(ppg.NotADag):
            jobA.depends_on(jobB)

    def test_depends_on_accepts_a_list_of_lists(self):
        jobA = ppg.FileGeneratingJob("out/A", lambda of: write("out/A", "A"))
        jobB = ppg.FileGeneratingJob("out/B", lambda of: write("out/B", "B"))
        jobC = ppg.FileGeneratingJob(
            "out/C",
            lambda of: write("out/C", read("out/A") + read("out/B") + read("out/D")),
        )
        jobD = ppg.FileGeneratingJob("out/D", lambda of: write("out/D", "D"))
        jobC.depends_on([jobA, [jobB, jobD]])
        assert ppg.global_pipegraph.has_edge(jobD, jobC)
        assert ppg.global_pipegraph.has_edge(jobA, jobC)
        assert ppg.global_pipegraph.has_edge(jobB, jobC)
        ppg.run()
        assert read("out/A") == "A"
        assert read("out/B") == "B"
        assert read("out/C") == "ABD"
        assert read("out/D") == "D"

    def test_invariant_job_depends_on_raises(self):

        with pytest.raises(ppg.JobContractError):
            ppg.jobs._InvariantMixin().depends_on(ppg.Job(["B"]))

    def test_cached_job_depends_on(self):
        class Dummy:
            pass

        o = Dummy()
        jobA = ppg.CachedAttributeLoadingJob("cache/A", o, "a", lambda: 23)
        jobB = ppg.Job(["B"])
        jobC = ppg.Job(["C"])
        jobD = ppg.Job(["D"])
        jobA.calc.depends_on([jobB], jobC, jobD)
        assert not ppg.global_pipegraph.has_edge(jobB, jobA.load)
        assert not ppg.global_pipegraph.has_edge(jobC, jobA.load)
        assert not ppg.global_pipegraph.has_edge(jobD, jobA.load)
        assert ppg.global_pipegraph.has_edge(jobB, jobA.calc)
        assert ppg.global_pipegraph.has_edge(jobC, jobA.calc)
        assert ppg.global_pipegraph.has_edge(jobD, jobA.calc)

    def test_dependency_placeholder(self):
        jobA = ppg.FileGeneratingJob(
            "out/A", lambda of: write("out/A", "A" + read("out/B"))
        )
        jobB = ppg.FileGeneratingJob("out/B", lambda of: write("out/B", "B"))

        def gen_deps():
            logger.info("gen deps called")
            return [jobB]

        jobA.depends_on(gen_deps)
        ppg.run()
        assert read("out/A") == "AB"

    def test_dependency_placeholder2(self):
        jobA = ppg.FileGeneratingJob(
            "out/A", lambda of: write("out/A", "A" + read("out/B"))
        )

        def gen_deps():
            return ppg.FileGeneratingJob("out/B", lambda of: write("out/B", "B"))

        jobA.depends_on(gen_deps)
        ppg.run()
        assert read("out/A") == "AB"

    def test_dependency_placeholder_nested(self):
        jobA = ppg.FileGeneratingJob(
            "out/A", lambda of: write("out/A", "A" + read("out/B") + read("out/C"))
        )

        def gen_deps2():
            return ppg.FileGeneratingJob("out/C", lambda of: write("out/C", "C"))

        def gen_deps():
            return ppg.FileGeneratingJob(
                "out/B", lambda of: write("out/B", "B")
            ).depends_on(gen_deps2)

        jobA.depends_on(gen_deps)
        ppg.run()
        assert read("out/A") == "ABC"

    def test_dependency_placeholder_dynamic_auto_invariants(self):
        jobA = ppg.FileGeneratingJob(
            "out/A", lambda of: write("out/A", "A" + read("out/B"))
        )

        def check_function_invariant(of):
            write("out/B", "B")
            assert (
                "FITestDependency.test_dependency_placeholder_dynamic_auto_invariants.<locals>.check_function_invariant"
                in ppg.global_pipegraph.jobs
            )

        def gen_deps():
            jobB = ppg.FileGeneratingJob("out/B", check_function_invariant)
            print("gen deps called")
            return [jobB]

        jobA.depends_on(gen_deps)
        assert "FIout/B" not in ppg.global_pipegraph.jobs
        ppg.run()
        assert read("out/A") == "AB"


@pytest.mark.usefixtures("ppg2_per_test")
class TestDefinitionErrors:
    def test_defining_function_invariant_twice(self):
        a = lambda: 55  # noqa:E731
        b = lambda: 66  # noqa:E731
        ppg.FunctionInvariant("a", a)
        ppg.FunctionInvariant("a", a)  # that's ok...

        with pytest.raises(ppg.JobRedefinitionError):
            ppg.FunctionInvariant("a", b)

        ppg.new(run_mode=ppg.RunMode.NOTEBOOK)
        ppg.FunctionInvariant("a", a)
        j = ppg.FunctionInvariant("a", b)
        assert j.function is b

    def test_defining_function_and_parameter_invariant_with_same_name(self):
        # you can't really, FunctionInvariants are Prefixed with FI, ParameterInvariant with PI
        a = lambda: 55  # noqa:E731
        ppg.FunctionInvariant("PIa", a)
        ppg.ParameterInvariant("a", "b")

    def test_defining_function_and_parameter_invariant_with_same_name_reversed(self):
        a = lambda: 55  # noqa:E731
        ppg.ParameterInvariant("a", "b")
        ppg.FunctionInvariant("PIa", a)

    def test_parameter_invariant_does_not_accept_function(self):
        with pytest.raises(TypeError):
            ppg.ParameterInvariant("a", lambda: 55)


@pytest.mark.usefixtures("ppg2_per_test")
class TestFunctionInvariantDisChanges_BetweenVersions:
    def test_function_name_is_irrelevant(self):
        def test_a():
            return 55

        def test_b():
            return 55

        def test_c():
            return 56

        a = ppg.FunctionInvariant("a", test_a)
        b = ppg.FunctionInvariant("b", test_b)
        c = ppg.FunctionInvariant("c", test_c)
        av = a.run(None, None)
        bv = b.run(None, None)
        cv = c.run(None, None)
        assert ppg.FunctionInvariant.compare_hashes(
            None, first_value(av), first_value(bv)
        )
        assert not ppg.FunctionInvariant.compare_hashes(
            None, first_value(av), first_value(cv)
        )

    def test_docstring_is_irrelevant(self):
        def test():
            """A"""
            return 55

        a = ppg.FunctionInvariant("a", test)

        # fmt: off
        def test():
            '''B'''
            return 55
        # fmt: on
        b = ppg.FunctionInvariant("b", test)

        def test():
            "c"
            return 56

        c = ppg.FunctionInvariant("c", test)

        def test():
            "c"
            return 56

        d = ppg.FunctionInvariant("d", test)
        av = first_value(a.run(None, None))
        bv = first_value(b.run(None, None))
        cv = first_value(c.run(None, None))
        dv = first_value(d.run(None, None))
        assert ppg.FunctionInvariant.compare_hashes(None, (av), (bv))
        assert ppg.FunctionInvariant.compare_hashes(None, (cv), (dv))
        assert not ppg.FunctionInvariant.compare_hashes(None, (av), (cv))



from pathlib import Path


class Dummy(object):
    pass


def write(filename, text):
    Path(filename).write_text(text)


def append(filename, text):
    p = Path(filename)
    if p.exists():
        old = p.read_text()
    else:
        old = ""
    p.write_text(old + text)


def writeappend(filename_write, filename_append, string):
    write(filename_write, string)
    append(filename_append, string)


def read(filename):
    return Path(filename).read_text()


def counter(filename):
    """Helper for counting invocations in a side-effect file"""
    try:
        res = int(Path(filename).read_text())
    except:  # noqa: E722
        res = 0
    Path(filename).write_text(str(res + 1))
    return str(res)


def force_load(job):
    """Force the loading of a Dataloading job that has no other dependents"""
    import pypipegraph2 as ppg

    ppg.JobGeneratingJob(job.job_id + "_gen", lambda: None).depends_on(job)



# check if 'stalling' occurs - ie.
# will having to many all_cores jobs ready to run
# prevent all single cores to run from in parallel
import pypipegraph2 as ppg
import subprocess
import time
from pathlib import Path
import logging
import os
import shutil
import sys

p = Path("run/call_externals")
if p.exists():
    shutil.rmtree(p)
p.mkdir(exist_ok=True, parents=True)
os.chdir(p)

timeout = 1
jobcount = 10
ppg.new(log_level=logging.INFO, cores=5)


def single_cores(ii):
    def inner(of):
        pid = os.fork()
        if pid == 0:
            of.write_bytes(
                subprocess.check_output(f"sleep {timeout} && date", shell=True)
            )
            sys.stdout.close()
            sys.stderr.close()
            time.sleep(1)
        else:
            os.waitpid(pid, 0)

        # time.sleep(timeout)

    return ppg.FileGeneratingJob(
        f"single{ii}",
        inner,
        resources=ppg.Resources.SingleCore,
        depend_on_function=False,
    )


sc = [single_cores(ii) for ii in range(jobcount)]

start = time.time()
ppg.run()
stop = time.time()

print(f"took {stop-start:.2f}.")



"""
# todo: rename file.
"""

from .enums import (
    JobOutcome,
)


class RecordedJobOutcome:
    """Job run information collector"""

    def __init__(self, job_id, outcome, payload):
        if not isinstance(outcome, JobOutcome):
            raise ValueError("Not an JobOutcome")
        self.job_id = job_id
        self.outcome = outcome
        self.payload = payload
        self.runtime = -1

    @property
    def error(self):
        if self.outcome is JobOutcome.Failed:
            return self.payload
        elif self.outcome is JobOutcome.UpstreamFailed:
            return "Upstream"
        else:
            raise AttributeError(
                f"No error available on an non failed JobOutcome (was {self.outcome}"
            )

    def __repr__(self):
        return f"RecordedJobOutcome({self.job_id}, {self.outcome}"



import pytest
import pypipegraph2 as ppg
from .shared import write


@pytest.mark.usefixtures("ppg2_per_test")
class TestCycles:
    def test_simple_cycle(self):
        with pytest.raises(ppg.exceptions.NotADag):
            jobA = ppg.FileGeneratingJob("A", lambda of: write("A", "A"))
            jobB = ppg.FileGeneratingJob("B", lambda of: write("B", "A"))
            jobA.depends_on(jobB)
            jobB.depends_on(jobA)
            # ppg.run()

    def test_indirect_cicle(self):
        jobA = ppg.FileGeneratingJob("A", lambda of: write("A", "A"))
        jobB = ppg.FileGeneratingJob("B", lambda of: write("B", "A"))
        jobC = ppg.FileGeneratingJob("C", lambda of: write("C", "A"))
        jobC.depends_on(jobB)
        jobB.depends_on(jobA)
        jobA.depends_on(jobC)

        with pytest.raises(ppg.exceptions.NotADag):
            ppg.run()

    def test_exceeding_max_cycle(self):
        max_depth = 50
        # this raises
        jobs = []
        for x in range(0, max_depth - 1):
            j = ppg.FileGeneratingJob(str(x), lambda of: write(str(x), str(x)))
            if jobs:
                j.depends_on(jobs[-1])
            jobs.append(j)
        jobs[0].depends_on(j)

        with pytest.raises(ppg.exceptions.NotADag):
            ppg.run()

        ppg.new()
        jobs = []
        for x in range(0, max_depth + 100):
            j = ppg.FileGeneratingJob(str(x), lambda of: write(str(x), str(x)))
            if jobs:
                j.depends_on(jobs[-1])
            jobs.append(j)
        jobs[0].depends_on(j)

        with pytest.raises(ppg.exceptions.NotADag):
            ppg.run()



import sys
import pytest
import pypipegraph2 as ppg2
# from . import fixtures


def run():
    """a ppg/no ppg aware run wrapper"""
    if ppg2.inside_ppg():
        ppg2.run()
    else:
        pass


run_pipegraph = run  # ppg1 compatibility

fl_count = 0


def force_load(job, prefix=None):
    """make sure a dataloadingjob has been loaded (if applicable)"""
    if ppg2.inside_ppg():
        if not isinstance(job, ppg2.Job):
            if prefix is None:
                global fl_count
                fl_count += 1
                prefix = "fl_%i" % fl_count
        else:
            prefix = job.job_id
        return ppg2.JobGeneratingJob(prefix + "_force_load", lambda: None).depends_on(
            job
        )


class RaisesDirectOrInsidePipegraph(object):
    """Piece of black magic from the depths of _pytest
    that will check whether a piece of code will raise the
    expected expcition (if outside of ppg), or if it will
    raise the exception when the pipegraph is running

    Use as a context manager like pytest.raises"""

    def __init__(self, expected_exception, search_message=None):
        self.expected_exception = expected_exception
        self.message = "DID NOT RAISE {}".format(expected_exception)
        self.search_message = search_message
        self.excinfo = None

    def __enter__(self):
        import _pytest

        self.excinfo = object.__new__(_pytest._code.ExceptionInfo)
        return self.excinfo

    def __exit__(self, *tp):
        from _pytest.outcomes import fail

        if ppg2.inside_ppg():
            with pytest.raises(ppg2.RunFailed) as e:
                run()
            ex = e.value.exceptions[0]
            if isinstance(ex, ppg2.JobError):
                ex = ex.args[0]
            if not isinstance(ex, self.expected_exception):
                raise ValueError(
                    f"Unexpected exception. Expected {self.expected_exception}, found {e.value.exceptions[0]}"
                )

            if self.search_message:
                assert self.search_message in str(e.value.exceptions[0])
        else:
            __tracebackhide__ = True
            if tp[0] is None:
                fail(self.message)
            self.excinfo.__init__(tp)
            suppress_exception = issubclass(self.excinfo.type, self.expected_exception)
            if sys.version_info[0] == 2:
                raise ValueError("No python2 support")
            if suppress_exception:
                return True  # seep PEP_0343
            else:
                return False



from pathlib import Path
import pypipegraph2 as ppg
import time
import psutil
import shutil
import os
import subprocess
import setproctitle

print("stage 2 pid is ", os.getpid())

setproctitle.setproctitle("stage2")

print("the ppg runner pid is", os.getpid())
p = Path("shu_test")
if p.exists():
    shutil.rmtree(p)
p.mkdir()
os.chdir(p)

Path("out").mkdir()
ppg.new()


def do_a(of):
    Path("a.txt").write_text(f"from do_a {os.getpgid(0)}")
    setproctitle.setproctitle("do_a")
    subprocess.check_call(
        [Path(__file__).parent / "sleeper.sh", "ppg2_test_parallel_sentinel"]
    )


def kill_us():
    time.sleep(1)

    p = psutil.Process(os.getpid())
    ppg.util.log_error(f"now killing this jobs process. {os.getpid()}")
    p.kill()


jobA = ppg.FileGeneratingJob("A", do_a)
jobB = ppg.JobGeneratingJob("B", kill_us)

try:
    ppg.run()
except ppg.FatalGraphException as e:
    if "killed" in str(e):
        pass
    else:
        raise


print("all done")



from pathlib import Path
import pytest
import pypipegraph2 as ppg
from .shared import write, read


def forget_job_status(invariant_status_filename=None):
    """Delete the job status file - usually only useful for testing"""
    if invariant_status_filename is None:
        invariant_status_filename = ppg.global_pipegraph.get_history_filename()
    try:
        Path(invariant_status_filename).unlink()
    except OSError:
        pass


def destroy_global_pipegraph():
    """Free the current global pipegraph - usually only useful for testing"""
    ppg.global_pipegraph = None


@pytest.mark.usefixtures("ppg2_per_test")
class TestSimple:
    def test_job_creation_before_pipegraph_creation_raises(self):
        destroy_global_pipegraph()
        with pytest.raises(ValueError):
            ppg.FileGeneratingJob("A", lambda: None)

    def test_using_after_destruction(self):
        a = ppg.FileGeneratingJob("A", lambda of: None)
        destroy_global_pipegraph()
        with pytest.raises(ValueError):
            a.readd()

    def test_run_pipegraph_without_pipegraph_raises(self):
        destroy_global_pipegraph()
        with pytest.raises(ValueError):
            ppg.run()

    def test_can_run_twice(self):

        ppg.run()
        ppg.run()
        ppg.run()

    def test_can_add_jobs_after_run(self):

        ppg.new()
        ppg.run()
        ppg.FileGeneratingJob("A", lambda of: write(of, "A"))
        ppg.run()
        assert read("A") == "A"

    def test_non_default_status_filename(self):
        ppg.new(history_dir="shu")
        ppg.FileGeneratingJob("A", lambda of: write(of, "A"))
        ppg.run()
        assert (Path("shu") / ppg.global_pipegraph.get_history_filename().name).exists()
        assert not (
            Path(".ppg") / ppg.global_pipegraph.get_history_filename().name
        ).exists()
